"""Flat LLM settings from environment (single active provider for Normbase/runtime)."""
from __future__ import annotations

from pydantic import AliasChoices, Field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

_NO_API_KEY = frozenset({"ollama", "local", "embedding"})


class LLMSettings(BaseSettings):
    """One provider name + credentials; optional model overrides."""

    llm_provider: str = Field(
        validation_alias=AliasChoices("LLM_PROVIDER", "NORMBASE_LLM_PROVIDER"),
    )
    llm_api_key: str = Field(
        default="",
        validation_alias=AliasChoices("LLM_API_KEY", "NORMBASE_LLM_API_KEY"),
    )
    llm_model: str | None = Field(
        default=None,
        validation_alias=AliasChoices("LLM_MODEL", "NORMBASE_LLM_MODEL"),
    )
    llm_embedding_model: str | None = Field(
        default=None,
        validation_alias=AliasChoices(
            "LLM_EMBEDDING_MODEL", "NORMBASE_LLM_EMBEDDING_MODEL"
        ),
    )
    llm_base_url: str | None = Field(
        default=None,
        validation_alias=AliasChoices("LLM_BASE_URL", "NORMBASE_LLM_BASE_URL"),
    )
    llm_azure_endpoint: str | None = Field(
        default=None,
        validation_alias=AliasChoices(
            "LLM_AZURE_ENDPOINT", "NORMBASE_LLM_AZURE_ENDPOINT"
        ),
    )

    model_config = SettingsConfigDict(env_file=".env", extra="ignore")

    @model_validator(mode="after")
    def _require_credentials(self) -> LLMSettings:
        name = self.llm_provider.strip()
        if not name:
            raise ValueError("LLM_PROVIDER is empty")
        if name in _NO_API_KEY:
            return self
        if name == "azure":
            if not (self.llm_api_key or "").strip():
                raise ValueError("LLM_API_KEY is required for azure")
            if not (self.llm_azure_endpoint or "").strip():
                raise ValueError(
                    "LLM_AZURE_ENDPOINT is required for azure "
                    "(Azure resource URL, not the same as LLM_BASE_URL)"
                )
            return self
        if not (self.llm_api_key or "").strip():
            raise ValueError(f"LLM_API_KEY is required for provider {name!r}")
        return self


def settings_to_provider_block(s: LLMSettings) -> dict:
    """Build the config dict passed to create_provider."""
    block: dict = {}
    if (s.llm_api_key or "").strip():
        block["api_key"] = s.llm_api_key.strip()
    if s.llm_model and str(s.llm_model).strip():
        block["model"] = s.llm_model.strip()
    if s.llm_embedding_model and str(s.llm_embedding_model).strip():
        block["embedding_model"] = s.llm_embedding_model.strip()
    if s.llm_base_url and str(s.llm_base_url).strip():
        block["base_url"] = s.llm_base_url.strip()
    if s.llm_provider == "azure" and (s.llm_azure_endpoint or "").strip():
        block["endpoint"] = s.llm_azure_endpoint.strip()
    return block



"""AIF Project Configuration"""
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic import AliasChoices, Field, ValidationError


class Settings(BaseSettings):
    """Application Settings"""
    
    # Normbase persistence
    database_url: str = Field(
        validation_alias=AliasChoices("NORMBASE_DATABASE_URL", "DATABASE_URL")
    )
    
    # Normbase infra dependencies
    redis_url: str = Field(
        validation_alias=AliasChoices("NORMBASE_REDIS_URL", "REDIS_URL")
    )
    
    # Normbase vector store
    qdrant_url: str = Field(
        validation_alias=AliasChoices("NORMBASE_QDRANT_URL", "QDRANT_URL")
    )
    qdrant_collection: str = Field(
        validation_alias=AliasChoices("NORMBASE_QDRANT_COLLECTION", "QDRANT_COLLECTION")
    )
    qdrant_vector_size: int = Field(
        validation_alias=AliasChoices("NORMBASE_QDRANT_VECTOR_SIZE", "QDRANT_VECTOR_SIZE")
    )
    
    # Service listen/runtime
    api_host: str = Field(validation_alias=AliasChoices("NORMBASE_API_HOST", "API_HOST"))
    normbase_api_port: int = Field(validation_alias=AliasChoices("NORMBASE_API_PORT"))
    eda_api_port: int = Field(validation_alias=AliasChoices("EDA_API_PORT"))
    ui_port: int = Field(validation_alias=AliasChoices("UI_PORT"))
    api_debug: bool = Field(
        validation_alias=AliasChoices("NORMBASE_API_DEBUG", "API_DEBUG")
    )

    # Service discovery URLs
    normbase_base_url: str = Field(validation_alias=AliasChoices("NORMBASE_BASE_URL"))
    eda_base_url: str = Field(validation_alias=AliasChoices("EDA_BASE_URL"))
    ui_base_url: str = Field(validation_alias=AliasChoices("UI_BASE_URL"))

    # Shared security (read-only status exposure / downstream checks)
    aif_internal_service_token: str | None = Field(
        default=None,
        validation_alias=AliasChoices("AIF_INTERNAL_SERVICE_TOKEN"),
    )
    aif_api_key_salt: str | None = Field(
        default=None,
        validation_alias=AliasChoices("AIF_API_KEY_SALT"),
    )

    # Retrieval cache
    retrieve_cache_enabled: bool = Field(
        validation_alias=AliasChoices(
            "NORMBASE_RETRIEVE_CACHE_ENABLED", "RETRIEVE_CACHE_ENABLED"
        )
    )
    retrieve_cache_ttl_seconds: int = Field(
        validation_alias=AliasChoices(
            "NORMBASE_RETRIEVE_CACHE_TTL_SECONDS", "RETRIEVE_CACHE_TTL_SECONDS"
        )
    )
    retrieve_cache_prefix: str = Field(
        validation_alias=AliasChoices(
            "NORMBASE_RETRIEVE_CACHE_PREFIX", "RETRIEVE_CACHE_PREFIX"
        )
    )

    # Norm extraction (process JSON + uploads); paths relative to CWD unless absolute
    extraction_artifact_dir: str = Field(
        default="data/normbase/extraction/artifacts",
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_ARTIFACT_DIR", "EXTRACTION_ARTIFACT_DIR"
        ),
    )
    extraction_upload_dir: str = Field(
        default="data/normbase/extraction/uploads",
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_UPLOAD_DIR", "EXTRACTION_UPLOAD_DIR"
        ),
    )
    extraction_artifact_max_bytes: int = Field(
        default=33_554_432,  # 32 MiB
        ge=1024,
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_ARTIFACT_MAX_BYTES", "EXTRACTION_ARTIFACT_MAX_BYTES"
        ),
    )
    extraction_upload_max_bytes: int = Field(
        default=104_857_600,  # 100 MiB
        ge=1024,
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_UPLOAD_MAX_BYTES", "EXTRACTION_UPLOAD_MAX_BYTES"
        ),
    )
    extraction_upload_public_base_url: str | None = Field(
        default=None,
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_UPLOAD_PUBLIC_BASE_URL",
            "EXTRACTION_UPLOAD_PUBLIC_BASE_URL",
        ),
        description="Optional HTTPS base (no trailing slash) prepended to storage_relative_path for source_file_url.",
    )
    # LLM chunking / completion budgets for norm extraction live in ``norm_extractor.settings``
    # (env: NORMBASE_EXTRACTION_* or NORM_EXTRACTOR_* — see norm_extractor/settings.py and .env.example.normbase).

    model_config = SettingsConfigDict(
        env_file=".env",
        extra="allow",
    )

def _load_settings() -> Settings:
    try:
        return Settings()
    except ValidationError as exc:
        missing = [
            ".".join(str(p) for p in err.get("loc", []))
            for err in exc.errors()
            if err.get("type") == "missing"
        ]
        if missing:
            raise RuntimeError(
                "Missing required environment variables: "
                + ", ".join(missing)
            ) from exc
        raise RuntimeError(f"Invalid environment configuration: {exc}") from exc


settings = _load_settings()



"""Load norm extraction system prompt from package data."""
from __future__ import annotations

import re
from pathlib import Path

_PROMPT_PATH = Path(__file__).resolve().parent / "prompts" / "norm_extraction_prompt_zh.md"


def load_zh_extraction_system_prompt() -> str:
    """Return the system prompt block from prompts/norm_extraction_prompt_zh.md."""
    path = _PROMPT_PATH
    if not path.is_file():
        raise FileNotFoundError(f"Missing prompt file: {path}")
    raw = path.read_text(encoding="utf-8")
    m = re.search(
        r"## System Prompt[^\n]*\n\n(你是 AIF[\s\S]+?)\n---\n\n## User Prompt",
        raw,
    )
    if not m:
        raise ValueError(
            "Could not parse system prompt block in norm_extraction_prompt_zh.md"
        )
    return m.group(1).strip()



"""EDA Agent API routes."""
from __future__ import annotations

from typing import List
import uuid

from fastapi import APIRouter, Depends, Query, Request, status
from sqlalchemy.ext.asyncio import AsyncSession

from eda_agent.engine import EDAEngine
from eda_agent.api.deps_auth import ensure_agent_header_match
from eda_agent.db.connection import get_db
from eda_agent.models.persistence import AgentProfile
from eda_agent.schemas import (
    AgentCreateRequest,
    AgentListItemResponse,
    AgentUpdateRequest,
    AgentStateResponse,
    AgentSummaryResponse,
    BindNormsRequest,
    DecideRequest,
    DecisionResponse,
    ExecuteRequest,
    ExecutionResponse,
    PerceiveRequest,
    TraceResponse,
)
from eda_agent.services.agent_service import AgentService
from eda_agent.services.field_service import FieldService
from eda_agent.infra.normbase_client import HttpNormbaseClient
from eda_agent.infra.llm import get_llm_provider
from eda_agent.config import settings
from eda_agent.services.runtime_service import RuntimeService
from eda_agent.services.trace_service import TraceService

router = APIRouter(prefix="/agents", tags=["eda-agent"])

_field_service = FieldService()
_trace_service = TraceService()
_agent_service = AgentService(field_service=_field_service)


def _build_engine(_profile: AgentProfile) -> EDAEngine:
    provider = get_llm_provider()
    normbase = HttpNormbaseClient(
        base_url=settings.eda_normbase_base_url,
        retrieve_timeout_seconds=settings.eda_normbase_retrieve_timeout_seconds,
    )
    return EDAEngine(normbase_client=normbase, llm_provider=provider)


def _get_runtime_service() -> RuntimeService:
    # Keep runtime dependency overridable for tests via monkeypatch(_build_engine).
    return RuntimeService(
        agent_service=_agent_service,
        field_service=_field_service,
        trace_service=_trace_service,
        engine_builder=_build_engine,
    )


@router.post("", response_model=AgentSummaryResponse, status_code=status.HTTP_201_CREATED)
async def create_agent(
    request: AgentCreateRequest,
    db: AsyncSession = Depends(get_db),
):
    return await _agent_service.create_agent(db, request)


@router.get("", response_model=List[AgentListItemResponse])
async def list_agents(
    limit: int = Query(100, ge=1, le=500),
    db: AsyncSession = Depends(get_db),
):
    return await _agent_service.list_agents(db, limit=limit)


@router.get("/{agent_id}", response_model=AgentStateResponse)
async def get_agent_state(
    agent_id: uuid.UUID,
    raw_request: Request,
    db: AsyncSession = Depends(get_db),
):
    ensure_agent_header_match(agent_id, raw_request.headers.get("x-agent-id"))
    return await _agent_service.get_agent_state(db, agent_id=agent_id)


@router.patch("/{agent_id}", response_model=AgentSummaryResponse)
async def update_agent(
    agent_id: uuid.UUID,
    request: AgentUpdateRequest,
    raw_request: Request,
    db: AsyncSession = Depends(get_db),
):
    ensure_agent_header_match(agent_id, raw_request.headers.get("x-agent-id"))
    return await _agent_service.update_agent(db, agent_id=agent_id, request=request)


@router.delete("/{agent_id}", status_code=status.HTTP_204_NO_CONTENT)
async def delete_agent(
    agent_id: uuid.UUID,
    raw_request: Request,
    db: AsyncSession = Depends(get_db),
):
    ensure_agent_header_match(agent_id, raw_request.headers.get("x-agent-id"))
    await _agent_service.delete_agent(db, agent_id=agent_id)


@router.post("/{agent_id}/perceive", response_model=AgentStateResponse)
async def perceive(
    agent_id: uuid.UUID,
    request: PerceiveRequest,
    raw_request: Request,
    db: AsyncSession = Depends(get_db),
):
    ensure_agent_header_match(agent_id, raw_request.headers.get("x-agent-id"))
    return await _get_runtime_service().perceive(
        db,
        agent_id=agent_id,
        request=request,
        raw_request=raw_request,
    )


@router.post("/{agent_id}/bind_norms", response_model=AgentStateResponse)
async def bind_norms(
    agent_id: uuid.UUID,
    request: BindNormsRequest,
    raw_request: Request,
    db: AsyncSession = Depends(get_db),
):
    ensure_agent_header_match(agent_id, raw_request.headers.get("x-agent-id"))
    return await _get_runtime_service().bind_norms(
        db,
        agent_id=agent_id,
        request=request,
        raw_request=raw_request,
    )


@router.post("/{agent_id}/decide", response_model=DecisionResponse)
async def decide(
    agent_id: uuid.UUID,
    request: DecideRequest,
    raw_request: Request,
    db: AsyncSession = Depends(get_db),
):
    ensure_agent_header_match(agent_id, raw_request.headers.get("x-agent-id"))
    return await _get_runtime_service().decide(
        db,
        agent_id=agent_id,
        request=request,
        raw_request=raw_request,
    )


@router.post("/{agent_id}/execute", response_model=ExecutionResponse)
async def execute(
    agent_id: uuid.UUID,
    request: ExecuteRequest,
    raw_request: Request,
    db: AsyncSession = Depends(get_db),
):
    ensure_agent_header_match(agent_id, raw_request.headers.get("x-agent-id"))
    return await _get_runtime_service().execute(
        db,
        agent_id=agent_id,
        request=request,
        raw_request=raw_request,
    )


@router.get("/{agent_id}/traces", response_model=List[TraceResponse])
async def get_traces(
    agent_id: uuid.UUID,
    raw_request: Request,
    limit: int = Query(20, ge=1, le=200),
    db: AsyncSession = Depends(get_db),
):
    ensure_agent_header_match(agent_id, raw_request.headers.get("x-agent-id"))
    return await _get_runtime_service().get_traces(db, agent_id=agent_id, limit=limit)




"""AIF LLM Module - Shared LLM Providers"""
from llm.providers import (
    BaseLLMProvider,
    OpenAIProvider,
    AnthropicProvider,
    LLMProviderFactory,
    create_provider,
)
from llm.runtime import clear_llm_provider_cache, get_default_llm_provider
from llm.settings import LLMSettings, settings_to_provider_block

__all__ = [
    "BaseLLMProvider",
    "OpenAIProvider",
    "AnthropicProvider",
    "LLMProviderFactory",
    "create_provider",
    "LLMSettings",
    "settings_to_provider_block",
    "clear_llm_provider_cache",
    "get_default_llm_provider",
]



"""External-entry auth helpers for MCP adapter."""
from __future__ import annotations

import os

from security.auth.api_keys import ResolvedAPIKey
from security.auth.core import AuthInput, authorize, resolve_agent_api_key


def effective_caller_api_key(provided_api_key: str | None) -> str:
    """Resolve effective caller key with tool-arg precedence."""
    return (provided_api_key or "").strip() or os.getenv(
        "AIF_MCP_CALLER_API_KEY", ""
    ).strip()


async def authenticate_mcp_caller(provided_api_key: str | None) -> ResolvedAPIKey:
    """
    Authenticate MCP caller using per-agent keys only.

    Order:
    1) Tool arg `caller_api_key`
    2) Env fallback `AIF_MCP_CALLER_API_KEY`
    3) Resolve via Normbase `/api/v1/keys/resolve`
    """
    # Precedence keeps explicit tool params in control while supporting
    # client environments that cannot inject tool arguments automatically.
    effective_api_key = effective_caller_api_key(provided_api_key)
    if not effective_api_key:
        raise PermissionError("Unauthorized: missing MCP caller API key.")

    result = await authorize(
        AuthInput(
            entry_type="mcp",
            api_key=effective_api_key,
            action="tool:authenticate_mcp_caller",
            required_scopes=[],
        ),
        resolver=resolve_agent_api_key,
    )
    if result.decision != "allow":
        raise PermissionError("Unauthorized: invalid MCP API key.")
    return ResolvedAPIKey(
        key_id=result.key_id,
        agent_id=result.agent_id,
        scopes=result.scopes,
        is_active=True,
        expires_at=None,
    )


async def authorize_mcp_tool_call(
    provided_api_key: str | None,
    *,
    tool_name: str,
    required_scopes: list[str] | None = None,
    resource_agent_id: str | None = None,
) -> ResolvedAPIKey:
    """Authorize one MCP tool call and return normalized key principal."""
    effective_api_key = effective_caller_api_key(provided_api_key)
    result = await authorize(
        AuthInput(
            entry_type="mcp",
            api_key=effective_api_key,
            action=f"tool:{(tool_name or '').strip()}",
            required_scopes=list(required_scopes or []),
            resource_agent_id=resource_agent_id,
        ),
        resolver=resolve_agent_api_key,
    )
    if result.decision == "allow":
        return ResolvedAPIKey(
            key_id=result.key_id,
            agent_id=result.agent_id,
            scopes=result.scopes,
            is_active=True,
            expires_at=None,
        )
    if result.reason == "missing_api_key":
        raise PermissionError("Unauthorized: missing MCP caller API key.")
    if result.reason.startswith("missing_scope:"):
        scope = result.reason.split(":", 1)[1]
        raise PermissionError(f"Forbidden: missing required scope '{scope}'.")
    if result.reason == "agent_mismatch":
        raise PermissionError("Unauthorized: key does not match target agent.")
    raise PermissionError("Unauthorized: invalid MCP API key.")




"""Dashboard runtime configuration."""
from __future__ import annotations

from pydantic import AliasChoices, Field, ValidationError
from pydantic_settings import BaseSettings


class DashboardSettings(BaseSettings):
    host: str = Field(default="0.0.0.0", alias="DASHBOARD_HOST")
    port: int = Field(default=8010, alias="DASHBOARD_PORT")
    debug: bool = Field(default=False, alias="DASHBOARD_DEBUG")
    # Keep base URLs required (no defaults).
    normbase_base_url: str = Field(alias="NORMBASE_BASE_URL")
    eda_base_url: str = Field(alias="EDA_BASE_URL")
    mcp_base_url: str = Field(alias="MCP_BASE_URL")
    mcp_audit_log: str = Field(default="", alias="AIF_MCP_AUDIT_LOG")
    aif_internal_service_token: str = Field(default="", alias="AIF_INTERNAL_SERVICE_TOKEN")
    aif_api_key_salt: str = Field(default="", alias="AIF_API_KEY_SALT")

    # Neo4j (optional; used by dashboard graph viewer)
    neo4j_bolt_uri: str = Field(default="", alias="NEO4J_BOLT_URI")
    neo4j_user: str = Field(default="", alias="NEO4J_USER")
    neo4j_password: str = Field(default="", alias="NEO4J_PASSWORD")

    # Normbase extraction calls LLM (possibly many chunks); read timeout must exceed worst case.
    normbase_extraction_read_timeout_seconds: float = Field(
        default=900.0,
        ge=60.0,
        le=7200.0,
        validation_alias=AliasChoices(
            "DASHBOARD_NORMBASE_EXTRACTION_READ_TIMEOUT_SECONDS",
            "DASHBOARD_EXTRACTION_READ_TIMEOUT_SECONDS",
        ),
    )

    class Config:
        env_file = ".env"
        extra = "allow"


def _load_settings() -> DashboardSettings:
    try:
        settings = DashboardSettings()
    except ValidationError as exc:
        missing = [
            ".".join(str(part) for part in err.get("loc", []))
            for err in exc.errors()
            if err.get("type") == "missing"
        ]
        if missing:
            raise RuntimeError(
                "Missing required environment variables: " + ", ".join(missing)
            ) from exc
        raise RuntimeError(f"Invalid dashboard environment configuration: {exc}") from exc

    settings.normbase_base_url = settings.normbase_base_url.rstrip("/")
    settings.eda_base_url = settings.eda_base_url.rstrip("/")
    settings.mcp_base_url = settings.mcp_base_url.rstrip("/")
    settings.mcp_audit_log = settings.mcp_audit_log.strip()
    settings.aif_internal_service_token = settings.aif_internal_service_token.strip()
    settings.aif_api_key_salt = settings.aif_api_key_salt.strip()
    settings.neo4j_bolt_uri = settings.neo4j_bolt_uri.strip()
    settings.neo4j_user = settings.neo4j_user.strip()
    settings.neo4j_password = settings.neo4j_password.strip()
    return settings


settings = _load_settings()




"""EDA-first MCP server adapter over EDA runtime APIs."""
from __future__ import annotations

import argparse
from dataclasses import dataclass
from typing import Any, Dict, List
from uuid import uuid4
import os

from integrations.mcp.audit import write_audit_event
from integrations.mcp_eda import clients
from security.auth.core import AuthInput, AuthContext, authorize, resolve_agent_api_key

try:
    from mcp.server.fastmcp import FastMCP
except Exception as exc:  # pragma: no cover
    raise RuntimeError(
        "MCP SDK import failed. Install dependencies from requirements.txt first."
    ) from exc


mcp = FastMCP("aif-eda-governance")

_TOOL_SCOPE_MAP: dict[str, list[str]] = {
    "perceive": ["eda:perceive"],
    "bind_norms": ["eda:bind"],
    "decide": ["eda:decide"],
    "execute": ["eda:execute"],
    "get_traces": ["traces:read"],
}


def effective_caller_api_key(provided_api_key: str | None) -> str:
    """Resolve effective caller key with tool-arg precedence."""
    return (provided_api_key or "").strip() or os.getenv(
        "AIF_MCP_CALLER_API_KEY", ""
    ).strip()


@dataclass(frozen=True)
class _AuthorizedContext:
    auth: AuthContext
    caller_api_key: str


async def _authorize_tool_call(
    tool_name: str,
    caller_api_key: str | None,
    caller_id: str,
    request_id: str,
) -> _AuthorizedContext:
    required_scopes = _TOOL_SCOPE_MAP.get(tool_name, [])
    effective_api_key = effective_caller_api_key(caller_api_key)
    auth = await authorize(
        AuthInput(
            entry_type="mcp",
            api_key=effective_api_key,
            action=f"tool:{tool_name}",
            required_scopes=required_scopes,
            request_id=request_id,
            caller_id=caller_id,
        ),
        resolver=resolve_agent_api_key,
    )
    if auth.decision != "allow":
        if auth.reason == "missing_api_key":
            raise PermissionError("Unauthorized: missing MCP caller API key.")
        if auth.reason.startswith("missing_scope:"):
            scope = auth.reason.split(":", 1)[1]
            raise PermissionError(f"Forbidden: missing required scope '{scope}'.")
        raise PermissionError("Unauthorized: invalid MCP API key.")
    return _AuthorizedContext(auth=auth, caller_api_key=effective_api_key)


def _audit_policy_denied(request_id: str, caller_id: str, tool_name: str, error: str) -> None:
    write_audit_event(
        {
            "request_id": request_id,
            "caller_id": caller_id,
            "tool": tool_name,
            "ok": False,
            "meta": {
                "denied_reason": "auth_denied",
                "error": error,
            },
        }
    )


@mcp.tool()
async def perceive(
    input_data: Dict[str, Any],
    caller_id: str = "unknown",
    request_id: str = "",
    caller_api_key: str | None = None,
) -> Dict[str, Any]:
    """Perceive stage for the caller-bound EDA agent."""
    rid = request_id or f"mcp-eda-{uuid4()}"
    try:
        auth = await _authorize_tool_call("perceive", caller_api_key, caller_id, rid)
    except PermissionError as exc:
        _audit_policy_denied(rid, caller_id, "perceive", str(exc))
        raise
    result = await clients.perceive(
        agent_id=auth.auth.agent_id,
        input_data=input_data,
        request_id=rid,
        caller_id=caller_id,
        key_id=auth.auth.key_id,
    )
    write_audit_event(
        {
            "request_id": rid,
            "caller_id": caller_id,
            "tool": "perceive",
            "ok": True,
            "meta": {
                "agent_id": auth.auth.agent_id,
                "key_id": auth.auth.key_id,
                "belief_count": len((result.get("beliefs") or {})),
            },
        }
    )
    return result


@mcp.tool()
async def bind_norms(
    context: Dict[str, Any],
    caller_id: str = "unknown",
    request_id: str = "",
    caller_api_key: str | None = None,
) -> Dict[str, Any]:
    """Bind norms for the caller-bound EDA agent."""
    rid = request_id or f"mcp-eda-{uuid4()}"
    try:
        auth = await _authorize_tool_call("bind_norms", caller_api_key, caller_id, rid)
    except PermissionError as exc:
        _audit_policy_denied(rid, caller_id, "bind_norms", str(exc))
        raise
    result = await clients.bind_norms(
        agent_id=auth.auth.agent_id,
        context=context,
        request_id=rid,
        caller_id=caller_id,
        key_id=auth.auth.key_id,
    )
    write_audit_event(
        {
            "request_id": rid,
            "caller_id": caller_id,
            "tool": "bind_norms",
            "ok": True,
            "meta": {
                "agent_id": auth.auth.agent_id,
                "key_id": auth.auth.key_id,
                "norm_binding_count": len(result.get("norm_bindings", []) or []),
            },
        }
    )
    return result


@mcp.tool()
async def decide(
    candidate_actions: List[Dict[str, Any]],
    caller_id: str = "unknown",
    request_id: str = "",
    caller_api_key: str | None = None,
) -> Dict[str, Any]:
    """Decide stage for the caller-bound EDA agent."""
    rid = request_id or f"mcp-eda-{uuid4()}"
    try:
        auth = await _authorize_tool_call("decide", caller_api_key, caller_id, rid)
    except PermissionError as exc:
        _audit_policy_denied(rid, caller_id, "decide", str(exc))
        raise
    result = await clients.decide(
        agent_id=auth.auth.agent_id,
        candidate_actions=candidate_actions,
        request_id=rid,
        caller_id=caller_id,
        key_id=auth.auth.key_id,
    )
    write_audit_event(
        {
            "request_id": rid,
            "caller_id": caller_id,
            "tool": "decide",
            "ok": True,
            "meta": {
                "agent_id": auth.auth.agent_id,
                "key_id": auth.auth.key_id,
                "candidate_count": len(candidate_actions or []),
            },
        }
    )
    return result


@mcp.tool()
async def execute(
    action: Dict[str, Any],
    caller_id: str = "unknown",
    request_id: str = "",
    caller_api_key: str | None = None,
) -> Dict[str, Any]:
    """Execute stage for the caller-bound EDA agent."""
    rid = request_id or f"mcp-eda-{uuid4()}"
    try:
        auth = await _authorize_tool_call("execute", caller_api_key, caller_id, rid)
    except PermissionError as exc:
        _audit_policy_denied(rid, caller_id, "execute", str(exc))
        raise
    result = await clients.execute(
        agent_id=auth.auth.agent_id,
        action=action,
        request_id=rid,
        caller_id=caller_id,
        key_id=auth.auth.key_id,
    )
    write_audit_event(
        {
            "request_id": rid,
            "caller_id": caller_id,
            "tool": "execute",
            "ok": True,
            "meta": {
                "agent_id": auth.auth.agent_id,
                "key_id": auth.auth.key_id,
                "success": bool(result.get("success", False)),
            },
        }
    )
    return result


@mcp.tool()
async def get_traces(
    limit: int = 20,
    caller_id: str = "unknown",
    request_id: str = "",
    caller_api_key: str | None = None,
) -> List[Dict[str, Any]]:
    """Read recent traces for the caller-bound EDA agent."""
    rid = request_id or f"mcp-eda-{uuid4()}"
    try:
        auth = await _authorize_tool_call("get_traces", caller_api_key, caller_id, rid)
    except PermissionError as exc:
        _audit_policy_denied(rid, caller_id, "get_traces", str(exc))
        raise
    result = await clients.get_traces(
        agent_id=auth.auth.agent_id,
        limit=limit,
        request_id=rid,
        caller_id=caller_id,
        key_id=auth.auth.key_id,
    )
    write_audit_event(
        {
            "request_id": rid,
            "caller_id": caller_id,
            "tool": "get_traces",
            "ok": True,
            "meta": {
                "agent_id": auth.auth.agent_id,
                "key_id": auth.auth.key_id,
                "limit": limit,
                "count": len(result or []),
            },
        }
    )
    return result


def main() -> None:
    """Run AIF EDA-first MCP adapter."""
    parser = argparse.ArgumentParser(description="Run AIF EDA-first MCP adapter.")
    parser.add_argument(
        "--transport",
        choices=("stdio", "sse", "streamable-http"),
        default="stdio",
        help="MCP transport mode (default: stdio)",
    )
    args = parser.parse_args()
    mcp.run(args.transport)


if __name__ == "__main__":
    main()



"""Norm extraction agent: text → LLM → raw norm dicts (no Normbase DB)."""
from __future__ import annotations

from norm_extractor.agent import RawExtractionOutcome, run_raw_extraction
from norm_extractor.exceptions import (
    BadExtractionJsonError,
    LlmExtractionError,
    NormExtractorError,
)

__all__ = [
    "BadExtractionJsonError",
    "LlmExtractionError",
    "NormExtractorError",
    "RawExtractionOutcome",
    "run_raw_extraction",
]



"""Business logic for norm CRUD and versioning."""
from __future__ import annotations

import uuid
from typing import Any, Dict, List, Optional

from fastapi import HTTPException, status
from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload

from normbase.config import settings
from normbase.models.norm import Norm, NormTag, NormVersion
from normbase.schemas import NormCreate, NormUpdate
from normbase.services.cache_service import CacheService
from normbase.services.norm_vector_service import NormVectorService


class NormService:
    """Service layer for norms."""
    BUSINESS_TAG_CATEGORIES = {
        "role",
        "scenario",
        "domain",
        "tool",
        "skill",
        "capability",
    }

    def __init__(self, vector_service: NormVectorService | None = None):
        self.vector_service = vector_service or NormVectorService()

    def _as_json(self, value: Any) -> Dict[str, Any]:
        """Convert typed schema payloads to plain JSON-serializable dicts."""
        if value is None:
            return {}
        if hasattr(value, "model_dump"):
            return value.model_dump()
        if isinstance(value, dict):
            return value
        return {}

    async def _invalidate_retrieve_cache(self) -> None:
        cache = CacheService()
        try:
            await cache.delete_by_prefix(settings.retrieve_cache_prefix)
        except Exception as e:
            print(f"Retrieve cache invalidation error: {e}")

    def _build_norm_snapshot(self, norm: Norm) -> Dict[str, Any]:
        return {
            "id": str(norm.id),
            "title": norm.title,
            "field_id": str(norm.field_id) if norm.field_id else None,
            "norm_type": norm.norm_type,
            "authority_level": norm.authority_level,
            "formality_level": norm.formality_level,
            "subject": norm.subject,
            "context": norm.context,
            "condition": norm.condition,
            "action": norm.action,
            "consequence": norm.consequence,
            "source": norm.source,
            "source_details": norm.source_details,
            "case_refs": norm.case_refs,
            "created_by": norm.created_by,
            "created_at": norm.created_at.isoformat() if norm.created_at else None,
            "updated_at": norm.updated_at.isoformat() if norm.updated_at else None,
            "status": norm.status,
            "version": norm.version,
            "tags": [{"tag": t.tag, "tag_category": t.tag_category} for t in norm.tags],
        }

    def _normalize_text_list(self, values: Any) -> List[str]:
        if isinstance(values, list):
            normalized = []
            for item in values:
                text = str(item).strip()
                if text:
                    normalized.append(text)
            return normalized
        return []

    def _canonical(self, payload: Any) -> Dict[str, Any]:
        if not isinstance(payload, dict):
            return {}
        canonical = payload.get("canonical", {})
        return canonical if isinstance(canonical, dict) else {}

    def _derive_business_tags(
        self,
        *,
        subject: Dict[str, Any],
        context: Dict[str, Any],
        action: Dict[str, Any],
    ) -> List[Dict[str, str]]:
        derived: List[Dict[str, str]] = []
        subject_canonical = self._canonical(subject)
        context_canonical = self._canonical(context)
        action_canonical = self._canonical(action)

        mapping = {
            "role": self._normalize_text_list(subject_canonical.get("roles", [])),
            "scenario": self._normalize_text_list(context_canonical.get("scenarios", [])),
            "domain": self._normalize_text_list(context_canonical.get("domains", [])),
            "capability": self._normalize_text_list(
                subject_canonical.get("capabilities", [])
            ),
            "tool": self._normalize_text_list(action_canonical.get("tools", [])),
            "skill": self._normalize_text_list(action_canonical.get("skills", [])),
        }
        for category, values in mapping.items():
            for tag in values:
                derived.append({"tag_category": category, "tag": tag})
        return derived

    def _normalize_input_tags(self, tags: Any) -> List[Dict[str, str]]:
        normalized: List[Dict[str, str]] = []
        if not isinstance(tags, list):
            return normalized
        for item in tags:
            if hasattr(item, "model_dump"):
                item = item.model_dump()
            if not isinstance(item, dict):
                continue
            category = str(item.get("tag_category", "")).strip().lower()
            tag = str(item.get("tag", "")).strip()
            if not category or not tag:
                continue
            normalized.append({"tag_category": category, "tag": tag})
        return normalized

    def _dedupe_tags(self, tags: List[Dict[str, str]]) -> List[Dict[str, str]]:
        deduped: List[Dict[str, str]] = []
        seen = set()
        for item in tags:
            key = (item["tag_category"], item["tag"])
            if key in seen:
                continue
            seen.add(key)
            deduped.append(item)
        return deduped

    def _merge_tags(
        self,
        *,
        derived_tags: List[Dict[str, str]],
        manual_tags: List[Dict[str, str]],
    ) -> List[Dict[str, str]]:
        filtered_manual = [
            item
            for item in manual_tags
            if item["tag_category"] not in self.BUSINESS_TAG_CATEGORIES
        ]
        return self._dedupe_tags([*derived_tags, *filtered_manual])

    async def create_norm(self, db: AsyncSession, norm_data: NormCreate) -> Norm:
        subject_payload = self._as_json(norm_data.subject)
        context_payload = self._as_json(norm_data.context)
        condition_payload = self._as_json(norm_data.condition)
        action_payload = self._as_json(norm_data.action)
        consequence_payload = self._as_json(norm_data.consequence)
        derived_tags = self._derive_business_tags(
            subject=subject_payload,
            context=context_payload,
            action=action_payload,
        )
        manual_tags = self._normalize_input_tags(norm_data.tags)
        merged_tags = self._merge_tags(derived_tags=derived_tags, manual_tags=manual_tags)

        norm = Norm(
            title=norm_data.title,
            field_id=norm_data.field_id,
            norm_type=norm_data.norm_type,
            authority_level=norm_data.authority_level,
            formality_level=norm_data.formality_level,
            subject=subject_payload,
            context=context_payload,
            condition=condition_payload,
            action=action_payload,
            consequence=consequence_payload,
            source=norm_data.source,
            source_details=norm_data.source_details,
            case_refs=norm_data.case_refs,
            created_by=norm_data.created_by,
            status=norm_data.status,
        )
        norm.tags = [
            NormTag(tag=item["tag"], tag_category=item["tag_category"])
            for item in merged_tags
        ]
        db.add(norm)
        await db.commit()
        result = await db.execute(
            select(Norm).options(selectinload(Norm.tags)).where(Norm.id == norm.id)
        )
        created = result.scalar_one()
        # Qdrant only holds active norms (NormVectorService.sync_norm also enforces this).
        if created.status == "active":
            await self.vector_service.sync_norm(created)
        await self._invalidate_retrieve_cache()
        return created

    async def list_norms(
        self,
        db: AsyncSession,
        *,
        page: int,
        page_size: int,
        norm_type: Optional[str],
        status_filter: Optional[str],
        authority_level: Optional[str],
        formality_level: Optional[str],
        field_id: Optional[uuid.UUID],
    ) -> tuple[int, List[Norm]]:
        query = select(Norm).options(selectinload(Norm.tags))
        count_query = select(func.count(Norm.id))

        if norm_type:
            query = query.where(Norm.norm_type == norm_type)
            count_query = count_query.where(Norm.norm_type == norm_type)
        if status_filter:
            query = query.where(Norm.status == status_filter)
            count_query = count_query.where(Norm.status == status_filter)
        if authority_level:
            query = query.where(Norm.authority_level == authority_level)
            count_query = count_query.where(Norm.authority_level == authority_level)
        if formality_level:
            query = query.where(Norm.formality_level == formality_level)
            count_query = count_query.where(Norm.formality_level == formality_level)
        if field_id:
            query = query.where(Norm.field_id == field_id)
            count_query = count_query.where(Norm.field_id == field_id)

        total_result = await db.execute(count_query)
        total = total_result.scalar() or 0
        query = query.offset((page - 1) * page_size).limit(page_size)
        result = await db.execute(query)
        norms = result.scalars().all()
        return total, list(norms)

    async def get_norm(self, db: AsyncSession, *, norm_id: uuid.UUID) -> Norm:
        result = await db.execute(
            select(Norm).options(selectinload(Norm.tags)).where(Norm.id == norm_id)
        )
        norm = result.scalar_one_or_none()
        if not norm:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Norm with id {norm_id} not found",
            )
        return norm

    async def update_norm(
        self,
        db: AsyncSession,
        *,
        norm_id: uuid.UUID,
        norm_data: NormUpdate,
    ) -> Norm:
        norm = await self.get_norm(db, norm_id=norm_id)

        version_snapshot = NormVersion(
            norm_id=norm.id,
            version=norm.version,
            snapshot=self._build_norm_snapshot(norm),
        )
        db.add(version_snapshot)

        update_data = norm_data.model_dump(exclude_unset=True)
        for field, value in update_data.items():
            if field == "tags":
                continue
            setattr(norm, field, value)
        should_recompute_tags = any(
            field in update_data for field in ("subject", "context", "action", "tags")
        )
        if should_recompute_tags:
            subject_payload = norm.subject if isinstance(norm.subject, dict) else {}
            context_payload = norm.context if isinstance(norm.context, dict) else {}
            action_payload = norm.action if isinstance(norm.action, dict) else {}
            derived_tags = self._derive_business_tags(
                subject=subject_payload,
                context=context_payload,
                action=action_payload,
            )
            manual_tags_source = update_data.get(
                "tags",
                [{"tag": t.tag, "tag_category": t.tag_category} for t in (norm.tags or [])],
            )
            manual_tags = self._normalize_input_tags(manual_tags_source)
            merged_tags = self._merge_tags(
                derived_tags=derived_tags,
                manual_tags=manual_tags,
            )
            norm.tags = [
                NormTag(tag=item["tag"], tag_category=item["tag_category"])
                for item in merged_tags
            ]

        norm.version += 1
        await db.commit()
        result = await db.execute(
            select(Norm).options(selectinload(Norm.tags)).where(Norm.id == norm.id)
        )
        updated = result.scalar_one()
        await self.vector_service.sync_norm(updated)
        await self._invalidate_retrieve_cache()
        return updated

    async def delete_norm(self, db: AsyncSession, *, norm_id: uuid.UUID) -> None:
        norm = await self.get_norm(db, norm_id=norm_id)
        norm.status = "archived"
        await db.commit()
        await self.vector_service.delete_norm(norm_id)
        await self._invalidate_retrieve_cache()

    async def list_norm_versions(
        self,
        db: AsyncSession,
        *,
        norm_id: uuid.UUID,
    ) -> List[NormVersion]:
        result = await db.execute(
            select(NormVersion)
            .where(NormVersion.norm_id == norm_id)
            .order_by(NormVersion.version.desc())
        )
        return list(result.scalars().all())

    async def create_norm_version(
        self,
        db: AsyncSession,
        *,
        norm_id: uuid.UUID,
        change_note: str,
    ) -> NormVersion:
        norm = await self.get_norm(db, norm_id=norm_id)
        version = NormVersion(
            norm_id=norm.id,
            version=norm.version,
            snapshot=self._build_norm_snapshot(norm),
            change_note=change_note,
        )
        db.add(version)
        await db.commit()
        await db.refresh(version)
        return version



"""AIF platform product boundary: `server` (Python backend) + `web` (user-facing frontend).

Package name is `aif_platform` (not `platform`) to avoid shadowing the Python standard library.

Prefer importing from `aif_platform.server.*` (e.g. `aif_platform.server.api.deps`).
"""



"""Business logic for API key lifecycle."""
from __future__ import annotations

import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import List, Optional

from fastapi import HTTPException
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from normbase.models.norm import APIKey
from security.auth.api_keys import generate_api_key, hash_api_key, verify_api_key_hash

DEFAULT_SCOPES = [
    "retrieval:plan",
    "norms:retrieve",
    "norms:check",
    "norms:report",
]


@dataclass
class IssuedKey:
    record: APIKey
    plain_key: str


class KeyService:
    """Service layer for API key operations."""

    def _utc_now(self) -> datetime:
        return datetime.now(timezone.utc).replace(tzinfo=None)

    def _to_utc(self, dt: datetime) -> datetime:
        if dt.tzinfo is None:
            return dt
        return dt.astimezone(timezone.utc).replace(tzinfo=None)

    def _ensure_not_expired(self, record: APIKey) -> bool:
        if record.expires_at is None:
            return True
        return self._to_utc(record.expires_at) > self._utc_now()

    def normalize_scopes(self, scopes: Optional[List[str]]) -> List[str]:
        normalized: List[str] = []
        for scope in scopes or []:
            text = str(scope).strip()
            if text and text not in normalized:
                normalized.append(text)
        return normalized or list(DEFAULT_SCOPES)

    def extract_bearer(self, authorization: Optional[str]) -> str:
        if not authorization:
            return ""
        raw = authorization.strip()
        if raw.lower().startswith("bearer "):
            return raw[7:].strip()
        return raw

    def agent_uuid_or_500(self, agent_id_text: Optional[str]) -> uuid.UUID:
        text = (agent_id_text or "").strip()
        if not text:
            raise HTTPException(status_code=500, detail="Stored API key missing agent_id UUID.")
        try:
            return uuid.UUID(text)
        except ValueError as exc:
            raise HTTPException(status_code=500, detail="Stored API key has invalid agent_id UUID.") from exc

    async def resolve_api_key_from_plain(
        self,
        db: AsyncSession,
        *,
        plain_api_key: str,
    ) -> APIKey:
        hashed = hash_api_key(plain_api_key)
        row = (
            await db.execute(select(APIKey).where(APIKey.key_hash == hashed))
        ).scalar_one_or_none()
        if not row:
            raise HTTPException(status_code=401, detail="Invalid API key")
        if not row.is_active or not self._ensure_not_expired(row):
            raise HTTPException(status_code=401, detail="API key inactive or expired")
        if not verify_api_key_hash(plain_api_key, row.key_hash):
            raise HTTPException(status_code=401, detail="Invalid API key")
        return row

    async def issue_api_key(
        self,
        db: AsyncSession,
        *,
        agent_id: uuid.UUID,
        name: Optional[str],
        expires_at: Optional[datetime],
        scopes: Optional[List[str]],
    ) -> IssuedKey:
        normalized_scopes = self.normalize_scopes(scopes)
        plain = generate_api_key()
        record = APIKey(
            key_hash=hash_api_key(plain),
            name=name or f"agent:{agent_id}",
            agent_id=str(agent_id),
            scopes=normalized_scopes,
            is_active=True,
            expires_at=expires_at,
        )
        db.add(record)
        await db.commit()
        await db.refresh(record)
        return IssuedKey(record=record, plain_key=plain)

    async def list_api_keys(
        self,
        db: AsyncSession,
        *,
        agent_id: Optional[uuid.UUID],
        include_inactive: bool,
    ) -> List[APIKey]:
        query = select(APIKey).order_by(APIKey.created_at.desc())
        if agent_id:
            query = query.where(APIKey.agent_id == str(agent_id))
        if not include_inactive:
            query = query.where(APIKey.is_active.is_(True))
        return list((await db.execute(query)).scalars().all())

    async def revoke_api_key(self, db: AsyncSession, *, key_id: uuid.UUID) -> APIKey:
        row = (await db.execute(select(APIKey).where(APIKey.id == key_id))).scalar_one_or_none()
        if not row:
            raise HTTPException(status_code=404, detail="API key not found")
        row.is_active = False
        await db.commit()
        await db.refresh(row)
        return row

    async def rotate_api_key(self, db: AsyncSession, *, key_id: uuid.UUID) -> IssuedKey:
        row = (await db.execute(select(APIKey).where(APIKey.id == key_id))).scalar_one_or_none()
        if not row:
            raise HTTPException(status_code=404, detail="API key not found")
        plain = generate_api_key()
        row.key_hash = hash_api_key(plain)
        row.is_active = True
        await db.commit()
        await db.refresh(row)
        return IssuedKey(record=row, plain_key=plain)



"""Business logic for role suggestion and retrieval planning."""
from __future__ import annotations

from collections import defaultdict
from typing import Any, Dict, List, Set

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from normbase.models.norm import Norm, NormTag
from normbase.schemas import RetrievalPlanResponse, RoleCandidate, RoleSuggestRequest

ROLE_CATEGORY = "role"
CONTEXT_TAG_CATEGORIES = ("scenario", "domain", "tool", "skill", "capability")
PLAN_TAG_CATEGORIES = (ROLE_CATEGORY, *CONTEXT_TAG_CATEGORIES)


def _flatten_context_terms(context: Dict[str, Any]) -> Set[str]:
    terms: Set[str] = set()

    def _walk(value: Any) -> None:
        if isinstance(value, dict):
            for k, v in value.items():
                terms.add(str(k).strip().lower())
                _walk(v)
            return
        if isinstance(value, list):
            for item in value:
                _walk(item)
            return
        if value is None:
            return
        terms.add(str(value).strip().lower())

    _walk(context or {})
    return {t for t in terms if t}


def _extract_context_tag_filters(context: Dict[str, Any]) -> Dict[str, List[str]]:
    filters: Dict[str, List[str]] = {}
    for category in CONTEXT_TAG_CATEGORIES:
        raw = (context or {}).get(category)
        values: List[str] = []
        if isinstance(raw, str):
            text = raw.strip()
            if text:
                values.append(text)
        elif isinstance(raw, list):
            for item in raw:
                text = str(item).strip()
                if text and text not in values:
                    values.append(text)
        if values:
            filters[category] = values
    return filters


def _empty_role_stats() -> Dict[str, Any]:
    return {
        "norm_count": 0,
        "matched_contexts": set(),
        "score": 0.0,
        "matched_tags": {category: set() for category in CONTEXT_TAG_CATEGORIES},
    }


class RoleService:
    """Service layer for role plan calculation."""

    async def build_role_plan(
        self,
        db: AsyncSession,
        request: RoleSuggestRequest,
    ) -> RetrievalPlanResponse:
        query = select(Norm).where(Norm.status == "active")
        if request.field_ids:
            query = query.where(Norm.field_id.in_(request.field_ids))

        norms = (await db.execute(query)).scalars().all()
        context_terms = _flatten_context_terms(request.context)
        role_stats: Dict[str, Dict[str, Any]] = defaultdict(_empty_role_stats)

        for norm in norms:
            tag_rows = (
                await db.execute(
                    select(NormTag).where(
                        NormTag.norm_id == norm.id,
                        NormTag.tag_category.in_(PLAN_TAG_CATEGORIES),
                    )
                )
            ).scalars().all()
            role_tags = [
                str(tag.tag).strip()
                for tag in tag_rows
                if tag.tag_category == ROLE_CATEGORY and str(tag.tag).strip()
            ]
            if not role_tags:
                continue

            matched_contexts = sorted(
                {
                    str(tag.tag).strip().lower()
                    for tag in tag_rows
                    if tag.tag_category in CONTEXT_TAG_CATEGORIES
                    and str(tag.tag).strip().lower() in context_terms
                }
            )

            matched_tags_by_category = {
                category: set() for category in CONTEXT_TAG_CATEGORIES
            }
            for tag in tag_rows:
                category = str(tag.tag_category).strip()
                if category not in matched_tags_by_category:
                    continue
                value = str(tag.tag).strip()
                if value and value.lower() in context_terms:
                    matched_tags_by_category[category].add(value)

            for role in role_tags:
                role_name = str(role).strip()
                if not role_name:
                    continue
                item = role_stats[role_name]
                item["norm_count"] += 1
                item["matched_contexts"].update(matched_contexts)
                for category in CONTEXT_TAG_CATEGORIES:
                    item["matched_tags"][category].update(matched_tags_by_category[category])
                base = 0.3 if role_name == "*" else 1.0
                item["score"] += base + (0.2 * len(matched_contexts))

        ranked = sorted(
            role_stats.items(),
            key=lambda pair: (pair[1]["score"], pair[1]["norm_count"]),
            reverse=True,
        )

        candidates: List[RoleCandidate] = []
        for role_name, item in ranked[: request.top_k]:
            candidates.append(
                RoleCandidate(
                    role=role_name,
                    score=round(float(item["score"]), 4),
                    norm_count=int(item["norm_count"]),
                    matched_contexts=sorted(item["matched_contexts"]),
                )
            )

        suggested_tag_filters = _extract_context_tag_filters(request.context)
        recommended_role = candidates[0].role if candidates else None
        role_is_global = recommended_role == "*"
        if recommended_role and recommended_role != "*":
            suggested_tag_filters[ROLE_CATEGORY] = [recommended_role]
        if role_is_global:
            suggested_tag_filters[ROLE_CATEGORY] = ["*"]
        if recommended_role and recommended_role in role_stats:
            matched_tags = role_stats[recommended_role]["matched_tags"]
            for category in CONTEXT_TAG_CATEGORIES:
                if category in suggested_tag_filters:
                    continue
                values = sorted(matched_tags.get(category, set()))
                if values:
                    suggested_tag_filters[category] = values

        return RetrievalPlanResponse(
            candidates=candidates,
            recommended_role=recommended_role,
            role_is_global=role_is_global,
            suggested_tag_filters=suggested_tag_filters,
        )



"""Plan Qdrant point set vs DB active norms (delete orphans, optional backfill)."""
from __future__ import annotations

from typing import FrozenSet, List, Tuple


def plan_norm_qdrant_reconcile(
    active_norm_ids: FrozenSet[str],
    qdrant_point_ids: FrozenSet[str],
    *,
    include_backfill: bool,
) -> Tuple[List[str], List[str]]:
    """
    Compute reconcile actions.

    Returns:
        to_delete: Qdrant point IDs to remove (not in active_norm_ids).
        to_backfill: Active norm IDs that have no Qdrant point (only if include_backfill).
    """
    to_delete = sorted(qdrant_point_ids - active_norm_ids)
    if not include_backfill:
        return to_delete, []
    to_backfill = sorted(active_norm_ids - qdrant_point_ids)
    return to_delete, to_backfill



"""Extractor-only settings (same env keys as normbase for one .env)."""
from __future__ import annotations

from pydantic import AliasChoices, Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class ExtractorSettings(BaseSettings):
    """LLM + chunking budgets used by run_raw_extraction."""

    extraction_llm_max_content_chars: int = Field(
        default=36_000,
        ge=4096,
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_LLM_MAX_CONTENT_CHARS",
            "EXTRACTION_LLM_MAX_CONTENT_CHARS",
            "NORM_EXTRACTOR_LLM_MAX_CONTENT_CHARS",
        ),
    )
    extraction_chunk_overlap_chars: int = Field(
        default=800,
        ge=0,
        le=50_000,
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_CHUNK_OVERLAP_CHARS",
            "EXTRACTION_CHUNK_OVERLAP_CHARS",
            "NORM_EXTRACTOR_CHUNK_OVERLAP_CHARS",
        ),
    )
    extraction_artifact_raw_llm_max_chars: int = Field(
        default=120_000,
        ge=10_000,
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_ARTIFACT_RAW_LLM_MAX_CHARS",
            "EXTRACTION_ARTIFACT_RAW_LLM_MAX_CHARS",
            "NORM_EXTRACTOR_ARTIFACT_RAW_LLM_MAX_CHARS",
        ),
    )
    extraction_llm_max_output_tokens: int = Field(
        default=32_768,
        ge=0,
        le=200_000,
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_LLM_MAX_OUTPUT_TOKENS",
            "EXTRACTION_LLM_MAX_OUTPUT_TOKENS",
            "NORM_EXTRACTOR_LLM_MAX_OUTPUT_TOKENS",
        ),
        description="max_tokens for completion; too low → truncated JSON. Provider hard limits apply (e.g. Volces Ark ≤32768).",
    )
    extraction_llm_temperature: float | None = Field(
        default=0.2,
        ge=0.0,
        le=2.0,
        validation_alias=AliasChoices(
            "NORMBASE_EXTRACTION_LLM_TEMPERATURE",
            "EXTRACTION_LLM_TEMPERATURE",
            "NORM_EXTRACTOR_LLM_TEMPERATURE",
        ),
    )

    model_config = SettingsConfigDict(env_file=".env", extra="ignore")


def get_extractor_settings() -> ExtractorSettings:
    return ExtractorSettings()



"""HTTP clients for Normbase and EDA services."""
from __future__ import annotations

import os
import uuid
from typing import Any, Dict, List, Literal

import httpx

from security.auth.verifier import build_internal_auth_headers


def _normbase_base_url() -> str:
    return os.getenv("NORMBASE_BASE_URL", "http://localhost:8001").rstrip("/")


def _eda_base_url() -> str:
    return os.getenv("EDA_BASE_URL", "http://localhost:8002").rstrip("/")


def _build_runtime_auth_headers(caller_api_key: str = "") -> dict[str, str]:
    key = (caller_api_key or "").strip()
    if not key:
        raise PermissionError("Unauthorized: missing MCP caller API key.")
    return {"X-API-Key": key}


def _normalize_agent_uuid(agent_id: str) -> str:
    text = (agent_id or "").strip()
    if not text:
        return ""
    return str(uuid.UUID(text))


async def retrieve_norms(
    context: Dict[str, Any],
    query_mode: Literal["structured", "semantic", "hybrid"] = "hybrid",
    nl_query: str | None = None,
    agent_role: str | None = None,
    agent_ids: List[str] | None = None,
    tag_filters: Dict[str, List[str]] | None = None,
    top_k: int = 20,
    field_ids: List[str] | None = None,
    request_id: str = "",
    caller_id: str = "",
    agent_id: str = "",
    key_id: str = "",
    caller_api_key: str = "",
) -> Dict[str, Any]:
    payload = {
        "context": context,
        "query_mode": query_mode,
        "top_k": top_k,
    }
    if nl_query:
        payload["nl_query"] = nl_query
    if agent_role:
        payload["agent_role"] = agent_role
    if agent_ids:
        payload["agent_ids"] = agent_ids
    if tag_filters:
        payload["tag_filters"] = tag_filters
    if field_ids:
        payload["field_ids"] = field_ids
    headers = _build_runtime_auth_headers(caller_api_key)
    if request_id:
        headers["X-Request-Id"] = request_id
    if caller_id:
        headers["X-Caller-Id"] = caller_id
    if agent_id:
        headers["X-Agent-Id"] = agent_id
    if key_id:
        headers["X-Key-Id"] = key_id
    async with httpx.AsyncClient(timeout=45.0) as client:
        resp = await client.post(
            f"{_normbase_base_url()}/api/v1/retrieve",
            json=payload,
            headers=headers,
        )
        resp.raise_for_status()
        return resp.json()


async def get_agent_field_ids(agent_id: str) -> List[str]:
    """Fetch EDA active_fields and keep only UUID-like field IDs."""
    normalized_agent_id = _normalize_agent_uuid(agent_id)
    if not normalized_agent_id:
        return []
    headers = build_internal_auth_headers()
    async with httpx.AsyncClient(timeout=20.0) as client:
        resp = await client.get(
            f"{_eda_base_url()}/api/v1/agents/{normalized_agent_id}",
            headers=headers,
        )
        resp.raise_for_status()
        data = resp.json() or {}
    raw_fields = data.get("active_fields") or []
    field_ids: List[str] = []
    for value in raw_fields:
        text = str(value).strip()
        if not text:
            continue
        try:
            field_ids.append(str(uuid.UUID(text)))
        except ValueError:
            continue
    return field_ids


async def plan_retrieval(
    context: Dict[str, Any],
    field_ids: List[str] | None = None,
    top_k: int = 5,
    request_id: str = "",
    caller_id: str = "",
    agent_id: str = "",
    key_id: str = "",
    caller_api_key: str = "",
) -> Dict[str, Any]:
    """Call Normbase retrieval planning endpoint."""
    payload: Dict[str, Any] = {
        "context": context,
        "top_k": top_k,
    }
    if field_ids:
        payload["field_ids"] = field_ids
    headers = _build_runtime_auth_headers(caller_api_key)
    if request_id:
        headers["X-Request-Id"] = request_id
    if caller_id:
        headers["X-Caller-Id"] = caller_id
    if agent_id:
        headers["X-Agent-Id"] = agent_id
    if key_id:
        headers["X-Key-Id"] = key_id
    async with httpx.AsyncClient(timeout=30.0) as client:
        resp = await client.post(
            f"{_normbase_base_url()}/api/v1/roles/plan",
            json=payload,
            headers=headers,
        )
        resp.raise_for_status()
        return resp.json()


async def get_traces(
    agent_id: str,
    limit: int = 20,
    request_id: str = "",
    caller_id: str = "",
    key_id: str = "",
) -> List[Dict[str, Any]]:
    headers = build_internal_auth_headers()
    if request_id:
        headers["X-Request-Id"] = request_id
    if caller_id:
        headers["X-Caller-Id"] = caller_id
    normalized_agent_id = _normalize_agent_uuid(agent_id)
    if normalized_agent_id:
        headers["X-Agent-Id"] = normalized_agent_id
    if key_id:
        headers["X-Key-Id"] = key_id
    async with httpx.AsyncClient(timeout=30.0) as client:
        resp = await client.get(
            f"{_eda_base_url()}/api/v1/agents/{normalized_agent_id}/traces",
            params={"limit": limit},
            headers=headers,
        )
        resp.raise_for_status()
        return resp.json()


async def check_action_compliance(
    usage_token: str,
    action_description: str,
    norm_ids: List[str] | None = None,
    request_id: str = "",
    caller_id: str = "",
    agent_id: str = "",
    key_id: str = "",
    caller_api_key: str = "",
) -> Dict[str, Any]:
    payload: Dict[str, Any] = {
        "usage_token": usage_token,
        "action_description": action_description,
    }
    if norm_ids:
        payload["norm_ids"] = norm_ids
    headers = _build_runtime_auth_headers(caller_api_key)
    if request_id:
        headers["X-Request-Id"] = request_id
    if caller_id:
        headers["X-Caller-Id"] = caller_id
    if agent_id:
        headers["X-Agent-Id"] = agent_id
    if key_id:
        headers["X-Key-Id"] = key_id
    async with httpx.AsyncClient(timeout=30.0) as client:
        resp = await client.post(
            f"{_normbase_base_url()}/api/v1/norms/check_action_compliance",
            json=payload,
            headers=headers,
        )
        resp.raise_for_status()
        return resp.json()


async def report_norm_outcome(
    usage_token: str,
    norm_id: str,
    outcome: str,
    details: Dict[str, Any] | None = None,
    request_id: str = "",
    caller_id: str = "",
    agent_id: str = "",
    key_id: str = "",
    caller_api_key: str = "",
) -> Dict[str, Any]:
    payload: Dict[str, Any] = {
        "usage_token": usage_token,
        "norm_id": norm_id,
        "outcome": outcome,
        "details": details or {},
    }
    if agent_id:
        payload["agent_id"] = agent_id
    headers = _build_runtime_auth_headers(caller_api_key)
    if request_id:
        headers["X-Request-Id"] = request_id
    if caller_id:
        headers["X-Caller-Id"] = caller_id
    if agent_id:
        headers["X-Agent-Id"] = agent_id
    if key_id:
        headers["X-Key-Id"] = key_id
    async with httpx.AsyncClient(timeout=30.0) as client:
        resp = await client.post(
            f"{_normbase_base_url()}/api/v1/norms/log-outcome",
            json=payload,
            headers=headers,
        )
        resp.raise_for_status()
        return resp.json()




"""Norms API Routes"""
import uuid
from typing import List, Optional
from fastapi import APIRouter, Depends, status, Query
from sqlalchemy.ext.asyncio import AsyncSession
from normbase.db.connection import get_db
from normbase.schemas import (
    NormCreate,
    NormUpdate,
    NormResponse,
    NormListResponse,
    NormVersionResponse,
)
from normbase.services.norm_service import NormService

router = APIRouter(prefix="/norms", tags=["norms"])
_norm_service = NormService()


@router.post("", response_model=NormResponse, status_code=status.HTTP_201_CREATED)
async def create_norm(
    norm_data: NormCreate,
    db: AsyncSession = Depends(get_db)
):
    """Create a new norm"""
    created = await _norm_service.create_norm(db, norm_data)
    return NormResponse.model_validate(created)


@router.get("", response_model=NormListResponse)
async def list_norms(
    page: int = Query(1, ge=1, description="Page number"),
    page_size: int = Query(20, ge=1, le=100, description="Page size"),
    norm_type: Optional[str] = Query(None, description="Filter by norm type"),
    status: Optional[str] = Query(None, description="Filter by status"),
    authority_level: Optional[str] = Query(None, description="Filter by authority level"),
    formality_level: Optional[str] = Query(None, description="Filter by formality level"),
    field_id: Optional[uuid.UUID] = Query(None, description="Filter by information field"),
    db: AsyncSession = Depends(get_db)
):
    """List all norms with pagination"""
    total, norms = await _norm_service.list_norms(
        db,
        page=page,
        page_size=page_size,
        norm_type=norm_type,
        status_filter=status,
        authority_level=authority_level,
        formality_level=formality_level,
        field_id=field_id,
    )
    return NormListResponse(
        total=total,
        items=[NormResponse.model_validate(n) for n in norms],
        page=page,
        page_size=page_size
    )


@router.get("/{norm_id}", response_model=NormResponse)
async def get_norm(
    norm_id: uuid.UUID,
    db: AsyncSession = Depends(get_db)
):
    """Get a norm by ID"""
    norm = await _norm_service.get_norm(db, norm_id=norm_id)
    return NormResponse.model_validate(norm)


@router.put("/{norm_id}", response_model=NormResponse)
async def update_norm(
    norm_id: uuid.UUID,
    norm_data: NormUpdate,
    db: AsyncSession = Depends(get_db)
):
    """Update a norm"""
    norm = await _norm_service.update_norm(db, norm_id=norm_id, norm_data=norm_data)
    return NormResponse.model_validate(norm)


@router.delete("/{norm_id}", status_code=status.HTTP_204_NO_CONTENT)
async def delete_norm(
    norm_id: uuid.UUID,
    db: AsyncSession = Depends(get_db)
):
    """Soft delete a norm by archiving it."""
    await _norm_service.delete_norm(db, norm_id=norm_id)


@router.get("/{norm_id}/versions", response_model=List[NormVersionResponse])
async def list_norm_versions(
    norm_id: uuid.UUID,
    db: AsyncSession = Depends(get_db)
):
    """Get version history for a norm"""
    versions = await _norm_service.list_norm_versions(db, norm_id=norm_id)
    return [NormVersionResponse.model_validate(v) for v in versions]


@router.post("/{norm_id}/versions", response_model=NormVersionResponse)
async def create_norm_version(
    norm_id: uuid.UUID,
    change_note: str,
    db: AsyncSession = Depends(get_db)
):
    """Create a new version snapshot for a norm"""
    version = await _norm_service.create_norm_version(
        db,
        norm_id=norm_id,
        change_note=change_note,
    )
    return NormVersionResponse.model_validate(version)



from __future__ import annotations

import argparse
import os
from pathlib import Path

from .loader import load_model
from .mapper import map_to_graph
from .neo4j_writer import Neo4jWriter


def _load_project_dotenv() -> None:
    """Load `.env` from the AIF repo root if python-dotenv is available."""
    try:
        from dotenv import load_dotenv
    except ImportError:
        return
    repo_root = Path(__file__).resolve().parents[2]
    load_dotenv(repo_root / ".env")


def build_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="archimate-to-graph",
        description=(
            "Import an ArchiMate model (Model Exchange Format) into a Neo4j graph.\n\n"
            "This tool is self-contained under `ontology/archimate_to_graph` and "
            "does not depend on or modify the rest of the AIF project."
        ),
    )
    parser.add_argument(
        "--input",
        required=True,
        help="Path to an ArchiMate model file (.archimate / .xml).",
    )
    parser.add_argument(
        "--uri",
        default=None,
        help="Neo4j Bolt URI. Default: NEO4J_BOLT_URI env, else bolt://127.0.0.1:7687.",
    )
    parser.add_argument(
        "--user",
        default=None,
        help="Neo4j username. Default: NEO4J_USER env, else neo4j.",
    )
    parser.add_argument(
        "--password",
        default=None,
        help="Neo4j password. Default: NEO4J_PASSWORD env, else empty string.",
    )
    return parser


def main(argv: list[str] | None = None) -> int:
    _load_project_dotenv()
    parser = build_arg_parser()
    parser.set_defaults(
        uri=os.getenv("NEO4J_BOLT_URI", "bolt://127.0.0.1:7687"),
        user=os.getenv("NEO4J_USER", "neo4j"),
        password=os.getenv("NEO4J_PASSWORD", ""),
    )
    args = parser.parse_args(argv)

    model_path = Path(args.input)
    if not model_path.exists():
        parser.error(f"Input file does not exist: {model_path}")

    model = load_model(model_path)
    nodes, edges = map_to_graph(model)

    writer = Neo4jWriter(args.uri, args.user, args.password)
    try:
        writer.import_graph(nodes, edges)
    finally:
        writer.close()

    return 0


if __name__ == "__main__":  # pragma: no cover
    raise SystemExit(main())




"""Business logic for information fields."""
from __future__ import annotations

from typing import List, Literal, Optional
from uuid import UUID

from fastapi import HTTPException
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from normbase.models.norm import Field as InfoField


class FieldService:
    """Service layer for information field operations."""

    async def list_fields(
        self,
        db: AsyncSession,
        status_filter: Optional[Literal["active", "inactive", "archived"]] = None,
    ) -> List[InfoField]:
        stmt = select(InfoField).order_by(InfoField.created_at.desc())
        if status_filter is not None:
            stmt = stmt.where(InfoField.status == status_filter)
        return (await db.execute(stmt)).scalars().all()

    async def create_field(
        self,
        db: AsyncSession,
        *,
        name: str,
        description: Optional[str],
        org_id: Optional[UUID],
        status_value: Literal["active", "inactive", "archived"],
    ) -> InfoField:
        normalized_name = (name or "").strip()
        if not normalized_name:
            raise HTTPException(status_code=400, detail="Field name cannot be empty.")

        existing_stmt = select(InfoField).where(InfoField.name == normalized_name)
        existing = (await db.execute(existing_stmt)).scalar_one_or_none()
        if existing:
            raise HTTPException(status_code=409, detail=f"Field '{normalized_name}' already exists.")

        item = InfoField(
            name=normalized_name,
            description=(description or None),
            org_id=org_id,
            status=status_value,
        )
        db.add(item)
        await db.commit()
        await db.refresh(item)
        return item

    async def update_field(
        self,
        db: AsyncSession,
        *,
        field_id: UUID,
        name: Optional[str],
        description: Optional[str],
        org_id: Optional[UUID],
        status_value: Optional[Literal["active", "inactive", "archived"]],
    ) -> InfoField:
        item = (
            await db.execute(select(InfoField).where(InfoField.id == field_id))
        ).scalar_one_or_none()
        if item is None:
            raise HTTPException(status_code=404, detail="Field not found.")

        if name is not None:
            normalized_name = name.strip()
            if not normalized_name:
                raise HTTPException(status_code=400, detail="Field name cannot be empty.")
            existing_stmt = select(InfoField).where(
                InfoField.name == normalized_name,
                InfoField.id != field_id,
            )
            existing = (await db.execute(existing_stmt)).scalar_one_or_none()
            if existing:
                raise HTTPException(status_code=409, detail=f"Field '{normalized_name}' already exists.")
            item.name = normalized_name

        if description is not None:
            item.description = description
        if org_id is not None:
            item.org_id = org_id
        if status_value is not None:
            item.status = status_value

        await db.commit()
        await db.refresh(item)
        return item

    async def soft_delete_field(self, db: AsyncSession, *, field_id: UUID) -> InfoField:
        item = (
            await db.execute(select(InfoField).where(InfoField.id == field_id))
        ).scalar_one_or_none()
        if item is None:
            raise HTTPException(status_code=404, detail="Field not found.")

        item.status = "archived"
        await db.commit()
        await db.refresh(item)
        return item



"""
archimate_to_graph
-------------------

Independent utility for importing ArchiMate models into Neo4j.

This module is intentionally isolated under `ontology/archimate_to_graph` and
does not modify or depend on the main AIF runtime modules.
"""

__all__ = ["loader", "mapper", "neo4j_writer"]




"""LLM norm extraction: chunk, call model, merge raw norm dicts (no DB / NormCreate)."""
from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from uuid import UUID

from norm_extractor.chunking import (
    dedupe_norm_dicts_by_title,
    distribute_max_norms_across_chunks,
    split_text_for_extraction,
)
from norm_extractor.exceptions import BadExtractionJsonError, LlmExtractionError
from norm_extractor.parsing import parse_llm_norms_json, truncate_raw_llm
from norm_extractor.prompts import load_zh_extraction_system_prompt
from norm_extractor.settings import get_extractor_settings

from llm.runtime import get_default_llm_provider

_JSON_RETRY_ATTEMPTS = 3


@dataclass(frozen=True)
class RawExtractionOutcome:
    """Immutable result of LLM extraction before NormCreate / DB."""

    norms_raw: List[Dict[str, Any]]
    chunk_results: List[Dict[str, Any]]
    chunk_count: int
    max_content_chars: int
    overlap_chars: int


async def _llm_extract_one_chunk(
    *,
    system_prompt: str,
    default_lang: str,
    field_id: Optional[UUID],
    created_by: Optional[str],
    chunk_max_norms: int,
    default_authority_level: Optional[str],
    default_formality_level: Optional[str],
    doc_id: str,
    doc_title: str,
    content: str,
    chunk_index: int,
    chunk_total: int,
    settings: Any,
) -> str:
    user_payload: Dict[str, Any] = {
        "task_mode": "extract_or_generate",
        "default_lang": default_lang,
        "target_field_id": str(field_id) if field_id else None,
        "default_created_by": created_by,
        "default_authority_level": default_authority_level,
        "default_formality_level": default_formality_level,
        "source_document": {
            "doc_id": doc_id,
            "title": doc_title,
            "version": "1",
            "language": default_lang,
            "content": content,
            "chunk_index": chunk_index,
            "chunk_total": chunk_total,
        },
        "extraction_scope": {
            "max_norms": chunk_max_norms,
            "include_permissions": True,
            "include_prohibitions": True,
            "include_obligations": True,
            "chunk_index": chunk_index,
            "chunk_total": chunk_total,
        },
    }
    user_message = (
        "请根据以下输入萃取并生成 norm 数据，严格遵循系统提示词中的字段合同与规则。\n\n"
        "输入：\n\n"
        f"{json.dumps(user_payload, ensure_ascii=False, indent=2)}\n\n"
        '输出要求：仅输出 JSON；顶层为 { "norms": [...] }；不要输出任何额外说明。'
    )
    llm = get_default_llm_provider()
    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": user_message},
    ]
    chat_kwargs: Dict[str, Any] = {}
    if settings.extraction_llm_max_output_tokens > 0:
        chat_kwargs["max_tokens"] = settings.extraction_llm_max_output_tokens
    if settings.extraction_llm_temperature is not None:
        chat_kwargs["temperature"] = settings.extraction_llm_temperature
    try:
        if chat_kwargs:
            try:
                return await llm.chat(messages, **chat_kwargs)
            except TypeError:
                return await llm.chat(messages)
        return await llm.chat(messages)
    except Exception as e:
        raise LlmExtractionError(str(e)) from e


async def run_raw_extraction(
    text: str,
    *,
    default_lang: str,
    field_id: Optional[UUID],
    created_by: Optional[str],
    max_norms: int,
    default_authority_level: Optional[str],
    default_formality_level: Optional[str],
    doc_id: str,
    doc_title: str,
) -> RawExtractionOutcome:
    """
    Run chunked LLM extraction; return raw norm dicts (not validated against NormCreate).

    Raises:
        LlmExtractionError: provider failure.
        BadExtractionJsonError: invalid JSON or missing ``norms`` array for a chunk.
    """
    settings = get_extractor_settings()
    system_prompt = load_zh_extraction_system_prompt()
    max_chars = settings.extraction_llm_max_content_chars
    overlap = settings.extraction_chunk_overlap_chars
    chunks = split_text_for_extraction(text or "", max_chars, overlap)
    per_chunk_limits = distribute_max_norms_across_chunks(max_norms, len(chunks))

    merged_norms_raw: List[Dict[str, Any]] = []
    chunk_results: List[Dict[str, Any]] = []
    raw_trunc_limit = settings.extraction_artifact_raw_llm_max_chars
    chunk_total = len(chunks)

    for i, chunk in enumerate(chunks):
        budget = per_chunk_limits[i]
        raw_reply = ""
        norms_part: List[Any] = []
        last_json_err: Exception | None = None
        requested_max = budget

        for attempt in range(_JSON_RETRY_ATTEMPTS):
            chunk_max = max(1, budget // (2**attempt))
            requested_max = chunk_max
            raw_reply = await _llm_extract_one_chunk(
                system_prompt=system_prompt,
                default_lang=default_lang,
                field_id=field_id,
                created_by=created_by,
                chunk_max_norms=chunk_max,
                default_authority_level=default_authority_level,
                default_formality_level=default_formality_level,
                doc_id=doc_id,
                doc_title=doc_title,
                content=chunk,
                chunk_index=i,
                chunk_total=chunk_total,
                settings=settings,
            )
            try:
                parsed = parse_llm_norms_json(raw_reply)
            except (json.JSONDecodeError, ValueError) as e:
                last_json_err = e
                if attempt + 1 >= _JSON_RETRY_ATTEMPTS:
                    raise BadExtractionJsonError(
                        f"LLM output is not valid JSON (chunk {i + 1}/{chunk_total}) "
                        f"after {_JSON_RETRY_ATTEMPTS} attempts "
                        f"(max_norms per try down to {chunk_max}): {e}"
                    ) from e
                continue

            np = parsed.get("norms")
            if not isinstance(np, list):
                last_json_err = ValueError("top-level 'norms' is not an array")
                if attempt + 1 >= _JSON_RETRY_ATTEMPTS:
                    raise BadExtractionJsonError(
                        f"LLM JSON must contain a 'norms' array (chunk {i + 1}/{chunk_total}) "
                        f"after {_JSON_RETRY_ATTEMPTS} attempts."
                    ) from last_json_err
                continue

            norms_part = np
            break

        cap = budget
        slice_norms = [
            x for x in norms_part[: cap + 10] if isinstance(x, dict)
        ][:cap]
        merged_norms_raw.extend(slice_norms)

        tr, was_trunc = truncate_raw_llm(raw_reply, raw_trunc_limit)
        chunk_results.append(
            {
                "index": i,
                "content_chars": len(chunk),
                "raw_llm": tr,
                "raw_llm_truncated": was_trunc,
                "norm_count": len(slice_norms),
                "max_norms_requested": requested_max,
                "json_parse_attempts": attempt + 1,
            }
        )

    norms_raw = dedupe_norm_dicts_by_title(merged_norms_raw)
    norms_raw = [x for x in norms_raw if isinstance(x, dict)][:max_norms]

    return RawExtractionOutcome(
        norms_raw=norms_raw,
        chunk_results=chunk_results,
        chunk_count=chunk_total,
        max_content_chars=max_chars,
        overlap_chars=overlap,
    )



"""Best-effort text extraction from uploaded bytes (PDF, plain text)."""
from __future__ import annotations

from typing import Optional, Tuple


def extract_text_from_upload(
    *,
    filename: str,
    content_type: Optional[str],
    data: bytes,
) -> Tuple[Optional[str], str, Optional[str]]:
    """
    Returns (extracted_text, parse_status, error_message).
    parse_status: done | skipped
    """
    safe = (filename or "").lower()
    ct = (content_type or "").lower()

    if "text/" in ct or safe.endswith(".txt") or safe.endswith(".md"):
        try:
            text = data.decode("utf-8")
            return text, "done", None
        except UnicodeDecodeError as e:
            return None, "skipped", f"utf-8 decode error: {e}"

    if safe.endswith(".pdf"):
        try:
            import fitz  # PyMuPDF

            doc = fitz.open(stream=data, filetype="pdf")
            try:
                parts: list[str] = []
                for page in doc:
                    t = page.get_text()
                    if t and t.strip():
                        parts.append(t.strip())
                merged = "\n\n".join(parts)
                if not merged.strip():
                    return None, "skipped", "pdf_no_extractable_text"
                return merged, "done", None
            finally:
                doc.close()
        except Exception as e:
            return None, "skipped", f"pdf_parse_error: {e}"[:500]

    return None, "skipped", None



"""Process-wide LLM client for Normbase (env-driven; see llm package)."""

from llm.runtime import get_default_llm_provider


def get_llm_provider():
    """Shared LLM from flat env: LLM_PROVIDER, LLM_API_KEY, LLM_MODEL, …"""
    try:
        return get_default_llm_provider()
    except Exception as e:
        raise RuntimeError(
            "Failed to resolve LLM provider for Normbase. "
            "Set LLM_PROVIDER / LLM_API_KEY / LLM_MODEL / LLM_EMBEDDING_MODEL / LLM_BASE_URL "
            "(see .env.example)."
        ) from e



"""Background extraction jobs (phase 4: async document extract)."""
from __future__ import annotations

from typing import Optional
from uuid import UUID

from normbase.db.connection import async_session_maker
from normbase.services.extraction_batch_service import ExtractionBatchService


async def run_batch_document_extraction_job(
    batch_id: UUID,
    document_id: UUID,
    *,
    default_lang: str,
    max_norms: int,
    submitter_user_id: Optional[str],
) -> None:
    """Run `extract_document` in a fresh DB session (e.g. after HTTP 202)."""
    async with async_session_maker() as db:
        batch_service = ExtractionBatchService()
        await batch_service.extract_document(
            db,
            batch_id=batch_id,
            document_id=document_id,
            default_lang=default_lang,
            max_norms=max_norms,
            submitter_user_id=submitter_user_id,
        )



"""Information fields management API."""
from __future__ import annotations

from typing import List, Optional, Literal
from uuid import UUID

from fastapi import APIRouter, Depends, Query, status
from sqlalchemy.ext.asyncio import AsyncSession

from normbase.db.connection import get_db
from normbase.schemas import FieldCreate, FieldResponse, FieldUpdate
from normbase.services.field_service import FieldService


router = APIRouter(tags=["fields"])
_field_service = FieldService()


@router.get("/fields", response_model=List[FieldResponse])
async def list_fields(
    status: Optional[Literal["active", "inactive", "archived"]] = Query(
        default=None,
        description="Filter by field status",
    ),
    db: AsyncSession = Depends(get_db),
):
    rows = await _field_service.list_fields(db, status_filter=status)
    return [FieldResponse.model_validate(row) for row in rows]


@router.post("/fields", response_model=FieldResponse, status_code=status.HTTP_201_CREATED)
async def create_field(payload: FieldCreate, db: AsyncSession = Depends(get_db)):
    item = await _field_service.create_field(
        db,
        name=payload.name,
        description=(payload.description or None),
        org_id=payload.org_id,
        status_value=payload.status,
    )
    return FieldResponse.model_validate(item)


@router.patch("/fields/{field_id}", response_model=FieldResponse)
async def update_field(field_id: UUID, payload: FieldUpdate, db: AsyncSession = Depends(get_db)):
    item = await _field_service.update_field(
        db,
        field_id=field_id,
        name=payload.name,
        description=payload.description,
        org_id=payload.org_id,
        status_value=payload.status,
    )
    return FieldResponse.model_validate(item)


@router.delete("/fields/{field_id}", response_model=FieldResponse)
async def soft_delete_field(field_id: UUID, db: AsyncSession = Depends(get_db)):
    item = await _field_service.soft_delete_field(db, field_id=field_id)
    return FieldResponse.model_validate(item)



"""Technical operations dashboard for AIF services (AIF Technical Dashboard)."""
from __future__ import annotations

from pathlib import Path
from typing import Any

import anyio
import httpx
from fastapi import FastAPI, File, HTTPException, UploadFile
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles

from dashboard.clients import DashboardClients
from dashboard.config import settings

app = FastAPI(title="AIF Technical Dashboard", version="0.1.0")
clients = DashboardClients()

BASE = Path(__file__).resolve().parent
app.mount("/static", StaticFiles(directory=str(BASE / "static")), name="static")


@app.get("/")
async def dashboard_index():
    return FileResponse(str(BASE / "templates" / "index.html"))


@app.get("/favicon.ico")
async def favicon():
    return FileResponse(str(BASE / "static" / "favicon.ico"))


@app.get("/health")
async def health():
    return {"status": "healthy", "name": "dashboard"}


def _request_error_detail(exc: httpx.RequestError) -> dict[str, Any]:
    """Map httpx transport errors to stable ``error_kind`` + specific hints."""
    url = str(exc.request.url) if exc.request else None
    base: dict[str, Any] = {
        "upstream_url": url,
        "error_type": type(exc).__name__,
    }
    tail = str(exc).strip() or None
    if tail:
        base["upstream_error_detail"] = tail[:2000]

    if isinstance(exc, httpx.ConnectTimeout):
        return {
            **base,
            "error_kind": "connect_timeout",
            "message": "Timed out while opening a connection to the upstream service.",
            "hint": "Upstream may be overloaded or blocked; check firewall and that the service is listening.",
        }
    if isinstance(exc, httpx.ReadTimeout):
        return {
            **base,
            "error_kind": "read_timeout",
            "message": "Upstream connected but did not send a complete response in time.",
            "hint": "The operation may still be running on the server; increase client timeout or retry. Long LLM calls need a high timeout.",
        }
    if isinstance(exc, httpx.WriteTimeout):
        return {
            **base,
            "error_kind": "write_timeout",
            "message": "Timed out while sending the request body to the upstream service.",
            "hint": "Try a smaller payload or check network stability.",
        }
    if isinstance(exc, httpx.PoolTimeout):
        return {
            **base,
            "error_kind": "pool_timeout",
            "message": "No free HTTP connection available to the upstream within the pool timeout.",
            "hint": "Reduce concurrent dashboard requests or increase httpx connection pool limits.",
        }
    if isinstance(exc, httpx.ConnectError):
        return {
            **base,
            "error_kind": "connect_error",
            "message": "Could not connect to the upstream host or port (connection refused or unreachable).",
            "hint": "Confirm the target process is running and NORMBASE_BASE_URL / port match (e.g. Normbase on 8001).",
        }
    if isinstance(exc, httpx.ReadError):
        return {
            **base,
            "error_kind": "read_error",
            "message": "Connection dropped while reading the response from upstream.",
            "hint": "Upstream may have restarted (e.g. uvicorn --reload) or closed the connection.",
        }
    if isinstance(exc, httpx.WriteError):
        return {
            **base,
            "error_kind": "write_error",
            "message": "Connection dropped while sending the request to upstream.",
            "hint": "Check network stability and whether the upstream process restarted.",
        }
    if isinstance(exc, httpx.RemoteProtocolError):
        return {
            **base,
            "error_kind": "remote_protocol_error",
            "message": "Upstream violated HTTP protocol or closed the connection unexpectedly.",
            "hint": "Often seen with proxies or mid-response crashes; check upstream logs.",
        }
    if isinstance(exc, httpx.LocalProtocolError):
        return {
            **base,
            "error_kind": "local_protocol_error",
            "message": "Local HTTP client hit a protocol error talking to upstream.",
            "hint": "Rare; note the upstream_error_detail and httpx version.",
        }
    if isinstance(exc, httpx.ProxyError):
        return {
            **base,
            "error_kind": "proxy_error",
            "message": "HTTP proxy rejected or failed the request.",
            "hint": "Check HTTP(S)_PROXY / NO_PROXY and corporate proxy settings.",
        }

    return {
        **base,
        "error_kind": "request_error",
        "message": "Request to upstream failed before an HTTP status was received.",
        "hint": "See error_type and upstream_error_detail; verify URL and that the service is reachable.",
    }


def _error(exc: Exception) -> HTTPException:
    if isinstance(exc, httpx.HTTPStatusError):
        status_code = exc.response.status_code
        upstream_url = str(exc.request.url)
        detail: dict[str, Any] = {
            "message": f"Upstream service returned HTTP {status_code}",
            "upstream_status": status_code,
            "upstream_url": upstream_url,
        }
        try:
            detail["upstream_body"] = exc.response.json()
        except Exception:
            detail["upstream_body"] = exc.response.text

        if status_code == 404 and "/api/v1/keys" in upstream_url:
            detail["hint"] = (
                "Normbase keys API route is missing. "
                "Restart Normbase with the latest code and verify /api/v1/keys is available."
            )
        elif status_code == 401:
            detail["hint"] = "Check internal token / API key configuration."
        elif status_code == 503:
            detail["hint"] = "Upstream service unavailable. Verify dependent services are running."
        elif status_code == 502 and "/extraction/" in upstream_url:
            ub = detail.get("upstream_body")
            dtext = ""
            if isinstance(ub, dict):
                raw = ub.get("detail")
                dtext = raw if isinstance(raw, str) else str(raw)
            elif isinstance(ub, str):
                dtext = ub
            if any(
                x in dtext
                for x in (
                    "not valid JSON",
                    "LLM output",
                    "Unterminated string",
                    "Expecting ",
                    "JSONDecodeError",
                )
            ):
                detail["hint"] = (
                    "Model output was cut off or malformed JSON. Norm_extractor already retries "
                    "with lower per-chunk max_norms (up to 3 calls). If it still fails: keep "
                    "NORMBASE_EXTRACTION_LLM_MAX_OUTPUT_TOKENS at your provider max (Ark 32768), "
                    "lower Dashboard max_norms, shorten text, or reduce "
                    "NORMBASE_EXTRACTION_LLM_MAX_CONTENT_CHARS to split into more chunks."
                )
                detail["error_kind"] = "upstream_extraction_llm_json"
            elif "max_tokens" in dtext and (
                "maximum" in dtext.lower() or "InvalidParameter" in dtext or "BadRequest" in dtext
            ):
                detail["hint"] = (
                    "max_tokens exceeds this LLM provider limit (e.g. Volces Ark ≤32768). "
                    "Set NORMBASE_EXTRACTION_LLM_MAX_OUTPUT_TOKENS within the allowed range."
                )
                detail["error_kind"] = "upstream_extraction_max_tokens"

        return HTTPException(status_code=502, detail=detail)

    if isinstance(exc, httpx.RequestError):
        detail = _request_error_detail(exc)
        return HTTPException(status_code=502, detail=detail)

    return HTTPException(status_code=500, detail={"message": str(exc)})


async def _ontology_neo4j_overview() -> dict[str, Any]:
    """Bolt connectivity for ontology graph viewer; failures do not fail the whole overview."""
    out: dict[str, Any] = {
        "configured": False,
        "reachable": None,
        "driver_installed": None,
    }
    if (
        not settings.neo4j_bolt_uri
        or not settings.neo4j_user
        or not settings.neo4j_password
    ):
        return out

    out["configured"] = True
    try:
        import neo4j  # noqa: F401
    except ImportError:
        out["driver_installed"] = False
        out["reachable"] = False
        return out

    out["driver_installed"] = True
    try:
        sync = await anyio.to_thread.run_sync(_neo4j_connectivity_sync)
    except Exception as exc:
        out["reachable"] = False
        out["error"] = str(exc)
        return out
    out["reachable"] = bool(sync.get("reachable"))
    if sync.get("error"):
        out["error"] = sync["error"]
    return out


def _neo4j_connectivity_sync() -> dict[str, Any]:
    from neo4j import GraphDatabase  # type: ignore

    driver = GraphDatabase.driver(
        settings.neo4j_bolt_uri, auth=(settings.neo4j_user, settings.neo4j_password)
    )
    try:
        driver.verify_connectivity()
        return {"reachable": True, "error": None}
    except Exception as exc:
        return {"reachable": False, "error": str(exc)}
    finally:
        driver.close()


@app.get("/api/overview")
async def api_overview():
    try:
        normbase = await clients.normbase_health()
        eda = await clients.eda_health()
        mcp_probe = await clients.mcp_health()
        mcp = {
            "probe": "live",
            "reachable": bool(mcp_probe.get("reachable")),
            "status_code": mcp_probe.get("status_code"),
        }
        recent_audit = clients.read_mcp_audit(limit=50)
        recent_errors = [
            item
            for item in recent_audit
            if item.get("ok") is False or item.get("status") in ("error", "failed")
        ][:10]
        ontology = {"neo4j": await _ontology_neo4j_overview()}
        return {
            "normbase": normbase,
            "eda": eda,
            "mcp": mcp,
            "ontology": ontology,
            "config": {
                "normbase_base_url": settings.normbase_base_url,
                "eda_base_url": settings.eda_base_url,
                "mcp_base_url": settings.mcp_base_url,
                "mcp_audit_log_configured": bool(settings.mcp_audit_log),
                "neo4j_bolt_uri": settings.neo4j_bolt_uri,
                "neo4j_user": settings.neo4j_user,
                "neo4j_password_configured": bool(settings.neo4j_password),
            },
            "security_config": {
                "internal_service_token_configured": bool(
                    (settings.aif_internal_service_token or "").strip()
                ),
                "api_key_salt_configured": bool(
                    (settings.aif_api_key_salt or "").strip()
                ),
            },
            "recent_errors": recent_errors,
        }
    except Exception as exc:
        raise _error(exc)


@app.get("/api/norms")
async def api_list_norms(
    page_size: int = 100,
    page: int = 1,
    status: str | None = None,
    field_id: str | None = None,
):
    try:
        return await clients.list_norms(
            page_size=page_size,
            page=page,
            status=status,
            field_id=field_id,
        )
    except Exception as exc:
        raise _error(exc)


@app.get("/api/norms/{norm_id}")
async def api_get_norm(norm_id: str):
    try:
        return await clients.get_norm(norm_id)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/norms")
async def api_create_norm(payload: dict[str, Any]):
    try:
        return await clients.create_norm(payload)
    except Exception as exc:
        raise _error(exc)


@app.put("/api/norms/{norm_id}")
async def api_update_norm(norm_id: str, payload: dict[str, Any]):
    try:
        return await clients.update_norm(norm_id, payload)
    except Exception as exc:
        raise _error(exc)


@app.delete("/api/norms/{norm_id}")
async def api_delete_norm(norm_id: str):
    try:
        return await clients.delete_norm(norm_id)
    except Exception as exc:
        raise _error(exc)


@app.get("/api/fields")
async def api_list_fields(status: str | None = None):
    try:
        return await clients.list_fields(status=status)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/fields")
async def api_create_field(payload: dict[str, Any]):
    try:
        return await clients.create_field(payload)
    except Exception as exc:
        raise _error(exc)


@app.patch("/api/fields/{field_id}")
async def api_update_field(field_id: str, payload: dict[str, Any]):
    try:
        return await clients.update_field(field_id, payload)
    except Exception as exc:
        raise _error(exc)


@app.delete("/api/fields/{field_id}")
async def api_soft_delete_field(field_id: str):
    try:
        return await clients.soft_delete_field(field_id)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/retrieve")
async def api_retrieve(payload: dict[str, Any]):
    try:
        return await clients.retrieve_norms(payload)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/compliance/check")
async def api_check_action_compliance(payload: dict[str, Any]):
    try:
        return await clients.check_action_compliance(payload)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/compliance/report")
async def api_report_norm_outcome(payload: dict[str, Any]):
    try:
        return await clients.report_norm_outcome(payload)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/extraction/from-text")
async def api_extraction_from_text(payload: dict[str, Any]):
    try:
        return await clients.extraction_from_text(payload)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/extraction/batches")
async def api_create_extraction_batch(payload: dict[str, Any]):
    try:
        return await clients.create_extraction_batch(payload)
    except Exception as exc:
        raise _error(exc)


@app.get("/api/extraction/batches/{batch_id}")
async def api_get_extraction_batch(batch_id: str):
    try:
        return await clients.get_extraction_batch(batch_id)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/extraction/batches/{batch_id}/documents")
async def api_upload_extraction_documents(
    batch_id: str,
    files: list[UploadFile] = File(...),
):
    try:
        parts: list[tuple[str, bytes, str | None]] = []
        for uf in files:
            data = await uf.read()
            parts.append((uf.filename or "upload.bin", data, uf.content_type))
        return await clients.upload_extraction_documents(batch_id, parts)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/extraction/batches/{batch_id}/documents/{document_id}/extract")
async def api_extract_extraction_document(
    batch_id: str,
    document_id: str,
    default_lang: str = "zh-CN",
    max_norms: int = 30,
    submitter_user_id: str | None = None,
):
    try:
        return await clients.extract_extraction_document(
            batch_id,
            document_id,
            default_lang=default_lang,
            max_norms=max_norms,
            submitter_user_id=submitter_user_id,
        )
    except Exception as exc:
        raise _error(exc)


@app.get("/api/agents")
async def api_list_agents(limit: int = 100):
    try:
        return await clients.list_agents(limit=limit)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/agents")
async def api_create_agent(payload: dict[str, Any]):
    try:
        return await clients.create_agent(payload)
    except Exception as exc:
        raise _error(exc)


@app.get("/api/agents/{agent_id}")
async def api_get_agent(agent_id: str):
    try:
        return await clients.get_agent(agent_id)
    except Exception as exc:
        raise _error(exc)


@app.patch("/api/agents/{agent_id}")
async def api_patch_agent(agent_id: str, payload: dict[str, Any]):
    try:
        return await clients.update_agent(agent_id, payload)
    except Exception as exc:
        raise _error(exc)


@app.delete("/api/agents/{agent_id}")
async def api_delete_agent(agent_id: str):
    try:
        return await clients.delete_agent(agent_id)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/agents/{agent_id}/{action}")
async def api_agent_action(agent_id: str, action: str, payload: dict[str, Any]):
    try:
        return await clients.agent_action(agent_id, action, payload)
    except Exception as exc:
        raise _error(exc)


@app.get("/api/agents/{agent_id}/traces")
async def api_get_traces(agent_id: str, limit: int = 20):
    try:
        return await clients.get_traces(agent_id, limit=limit)
    except Exception as exc:
        raise _error(exc)


@app.get("/api/mcp/info")
async def api_mcp_info():
    info: dict[str, Any] = {
        "tools": [],
        "mcp_base_url": settings.mcp_base_url,
        "auth_mode": "per-agent-key-only",
        "audit_log_configured": bool(settings.mcp_audit_log),
        "live": False,
    }
    try:
        info["tools"] = await clients.list_mcp_tools()
        info["live"] = True
    except Exception as exc:
        info["tools_error"] = str(exc)
    return info


def _neo4j_required() -> None:
    if not settings.neo4j_bolt_uri or not settings.neo4j_user or not settings.neo4j_password:
        raise HTTPException(
            status_code=400,
            detail={
                "message": "Neo4j is not configured for dashboard viewer.",
                "hint": "Set NEO4J_BOLT_URI, NEO4J_USER, NEO4J_PASSWORD in .env (or environment) and restart dashboard.",
            },
        )


def _fetch_neo4j_subgraph(limit: int, label: str, rel_type: str) -> dict[str, Any]:
    try:
        from neo4j import GraphDatabase  # type: ignore
    except ImportError as exc:
        raise HTTPException(
            status_code=500,
            detail={
                "message": "Neo4j driver is not installed.",
                "hint": "Install dependency `neo4j` and restart dashboard.",
            },
        ) from exc

    driver = GraphDatabase.driver(
        settings.neo4j_bolt_uri, auth=(settings.neo4j_user, settings.neo4j_password)
    )
    cypher = (
        "MATCH (n:`%s`)-[r:%s]->(m:`%s`) "
        "RETURN n, r, m "
        "LIMIT $limit"
    ) % (label, rel_type, label)

    nodes_by_id: dict[str, dict[str, Any]] = {}
    edges: list[dict[str, Any]] = []
    with driver.session() as session:
        result = session.run(cypher, limit=limit)
        for record in result:
            n = record["n"]
            m = record["m"]
            r = record["r"]

            for node in (n, m):
                nid = str(node.get("id") or node.element_id)
                if nid in nodes_by_id:
                    continue
                props = dict(node)
                props.setdefault("id", nid)
                nodes_by_id[nid] = {
                    "id": nid,
                    "labels": list(getattr(node, "labels", [])),
                    "properties": props,
                }

            edges.append(
                {
                    "id": str(r.get("id") or r.element_id),
                    # `r.type` in Neo4j driver is the relationship type token (e.g. "ARCHI_REL").
                    # The ArchiMate-specific relationship type we imported is stored as a property `type`.
                    "type": str(r.get("type") or getattr(r, "type", "") or rel_type),
                    "source": str(n.get("id") or n.element_id),
                    "target": str(m.get("id") or m.element_id),
                    "properties": dict(r),
                }
            )

    driver.close()
    return {"nodes": list(nodes_by_id.values()), "edges": edges, "cypher": cypher}


@app.get("/api/neo4j/subgraph")
async def api_neo4j_subgraph(
    limit: int = 200,
    label: str = "ArchimateElement",
    rel_type: str = "ARCHI_REL",
):
    _neo4j_required()
    safe_limit = max(1, min(int(limit), 1000))
    safe_label = (label or "").strip() or "ArchimateElement"
    safe_rel = (rel_type or "").strip() or "ARCHI_REL"

    # simple guard: allow only identifier-like tokens
    import re

    ident = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")
    if not ident.match(safe_label) or not ident.match(safe_rel):
        raise HTTPException(
            status_code=400,
            detail={
                "message": "Invalid label or relationship type.",
                "hint": "Use identifier tokens like ArchimateElement / ARCHI_REL (letters, numbers, underscore).",
            },
        )

    try:
        return await anyio.to_thread.run_sync(_fetch_neo4j_subgraph, safe_limit, safe_label, safe_rel)
    except HTTPException:
        raise
    except Exception as exc:
        raise HTTPException(
            status_code=502,
            detail={
                "message": "Failed to query Neo4j.",
                "hint": "Verify Neo4j is running and credentials are correct.",
                "error": str(exc),
            },
        )


@app.get("/api/mcp/audit")
async def api_mcp_audit(limit: int = 100):
    return {"items": clients.read_mcp_audit(limit=limit)}


@app.post("/api/mcp/rpc")
async def api_mcp_rpc(payload: dict[str, Any]):
    try:
        return await clients.mcp_rpc(payload)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/keys/issue")
async def api_issue_key(payload: dict[str, Any]):
    try:
        return await clients.issue_key(payload)
    except Exception as exc:
        raise _error(exc)


@app.get("/api/keys")
async def api_list_keys(agent_id: str | None = None, include_inactive: bool = False):
    try:
        return await clients.list_keys(agent_id=agent_id, include_inactive=include_inactive)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/keys/{key_id}/revoke")
async def api_revoke_key(key_id: str):
    try:
        return await clients.revoke_key(key_id)
    except Exception as exc:
        raise _error(exc)


@app.post("/api/keys/{key_id}/rotate")
async def api_rotate_key(key_id: str):
    try:
        return await clients.rotate_key(key_id)
    except Exception as exc:
        raise _error(exc)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(
        "dashboard.main:app",
        host=settings.host,
        port=settings.port,
        reload=settings.debug,
    )




"""API key management routes for integrations."""
from __future__ import annotations

import uuid
from typing import List, Optional

from fastapi import APIRouter, Depends, Header, HTTPException, Query, status
from sqlalchemy.ext.asyncio import AsyncSession

from normbase.api.deps_auth import RequireInternalToken
from normbase.db.connection import get_db
from normbase.schemas import (
    APIKeyIssueRequest,
    APIKeyIssueResponse,
    APIKeyListItem,
    APIKeyResolveRequest,
    APIKeyResolveResponse,
)
from security.auth.api_keys import mask_api_key
from normbase.services.key_service import KeyService

router = APIRouter(prefix="/keys", tags=["keys"])
_key_service = KeyService()


@router.post(
    "/issue",
    response_model=APIKeyIssueResponse,
    status_code=status.HTTP_201_CREATED,
    dependencies=[RequireInternalToken],
)
async def issue_api_key(
    request: APIKeyIssueRequest,
    db: AsyncSession = Depends(get_db),
):
    issued = await _key_service.issue_api_key(
        db,
        agent_id=request.agent_id,
        name=request.name,
        expires_at=request.expires_at,
        scopes=request.scopes,
    )
    record = issued.record
    plain = issued.plain_key
    return APIKeyIssueResponse(
        key_id=str(record.id),
        agent_id=_key_service.agent_uuid_or_500(record.agent_id),
        name=record.name,
        api_key=plain,
        key_preview=mask_api_key(plain),
        scopes=list(record.scopes or []),
        is_active=record.is_active,
        expires_at=record.expires_at,
        created_at=record.created_at,
    )


@router.get("", response_model=List[APIKeyListItem], dependencies=[RequireInternalToken])
async def list_api_keys(
    agent_id: Optional[uuid.UUID] = Query(None, description="Filter by agent_id (UUID)"),
    include_inactive: bool = Query(False),
    db: AsyncSession = Depends(get_db),
):
    rows = await _key_service.list_api_keys(
        db,
        agent_id=agent_id,
        include_inactive=include_inactive,
    )
    return [
        APIKeyListItem(
            key_id=str(item.id),
            agent_id=_key_service.agent_uuid_or_500(item.agent_id) if item.agent_id else None,
            name=item.name,
            key_preview="***",
            scopes=list(item.scopes or []),
            is_active=item.is_active,
            expires_at=item.expires_at,
            created_at=item.created_at,
        )
        for item in rows
    ]


@router.post("/{key_id}/revoke", response_model=APIKeyListItem, dependencies=[RequireInternalToken])
async def revoke_api_key(
    key_id: uuid.UUID,
    db: AsyncSession = Depends(get_db),
):
    row = await _key_service.revoke_api_key(db, key_id=key_id)
    return APIKeyListItem(
        key_id=str(row.id),
        agent_id=_key_service.agent_uuid_or_500(row.agent_id) if row.agent_id else None,
        name=row.name,
        key_preview="***",
        scopes=list(row.scopes or []),
        is_active=row.is_active,
        expires_at=row.expires_at,
        created_at=row.created_at,
    )


@router.post("/{key_id}/rotate", response_model=APIKeyIssueResponse, dependencies=[RequireInternalToken])
async def rotate_api_key(
    key_id: uuid.UUID,
    db: AsyncSession = Depends(get_db),
):
    issued = await _key_service.rotate_api_key(db, key_id=key_id)
    row = issued.record
    plain = issued.plain_key
    return APIKeyIssueResponse(
        key_id=str(row.id),
        agent_id=_key_service.agent_uuid_or_500(row.agent_id),
        name=row.name,
        api_key=plain,
        key_preview=mask_api_key(plain),
        scopes=list(row.scopes or []),
        is_active=row.is_active,
        expires_at=row.expires_at,
        created_at=row.created_at,
    )


@router.post("/resolve", response_model=APIKeyResolveResponse, dependencies=[RequireInternalToken])
async def resolve_api_key(
    request: APIKeyResolveRequest,
    db: AsyncSession = Depends(get_db),
):
    row = await _key_service.resolve_api_key_from_plain(
        db,
        plain_api_key=request.api_key,
    )
    return APIKeyResolveResponse(
        key_id=str(row.id),
        agent_id=_key_service.agent_uuid_or_500(row.agent_id),
        scopes=list(row.scopes or []),
        is_active=row.is_active,
        expires_at=row.expires_at,
    )


@router.get("/introspect", response_model=APIKeyResolveResponse)
async def introspect_api_key(
    authorization: Optional[str] = Header(default=None),
    x_api_key: Optional[str] = Header(default=None),
    db: AsyncSession = Depends(get_db),
):
    plain = (x_api_key or "").strip() or _key_service.extract_bearer(authorization)
    if not plain:
        raise HTTPException(status_code=401, detail="Missing API key")
    row = await _key_service.resolve_api_key_from_plain(db, plain_api_key=plain)
    return APIKeyResolveResponse(
        key_id=str(row.id),
        agent_id=_key_service.agent_uuid_or_500(row.agent_id),
        scopes=list(row.scopes or []),
        is_active=row.is_active,
        expires_at=row.expires_at,
    )




"""Extract norms from natural language via LLM; persist process JSON artifacts."""
from __future__ import annotations

import json
import uuid
from typing import Any, Dict, List, Optional, Tuple
from uuid import UUID

from fastapi import HTTPException, status
from pydantic import ValidationError
from sqlalchemy.ext.asyncio import AsyncSession

from norm_extractor import (
    BadExtractionJsonError,
    LlmExtractionError,
    run_raw_extraction,
)
from normbase.config import settings
from normbase.schemas import NormCreate
from normbase.services.extraction_artifact_storage import write_process_json
from normbase.services.norm_service import NormService

_NORM_CREATE_KEYS = frozenset(NormCreate.model_fields.keys())


def _coerce_norm_dict(raw: Dict[str, Any]) -> Dict[str, Any]:
    """Normalize LLM output toward NormCreate-compatible dict."""
    item = dict(raw)
    item["status"] = "draft"
    allowed = {k: v for k, v in item.items() if k in _NORM_CREATE_KEYS}
    fid = allowed.get("field_id")
    if fid == "":
        allowed["field_id"] = None
    subj = allowed.get("subject")
    if isinstance(subj, dict):
        can = subj.get("canonical")
        if isinstance(can, dict) and "agents" in can:
            agents = can.get("agents") or []
            parsed: List[Any] = []
            if isinstance(agents, list):
                for a in agents:
                    if isinstance(a, UUID):
                        parsed.append(a)
                    elif isinstance(a, str) and a.strip():
                        try:
                            parsed.append(UUID(a.strip()))
                        except ValueError:
                            continue
            can["agents"] = parsed
    if allowed.get("tags") is None:
        allowed["tags"] = []
    return allowed


def _validation_detail(exc: ValidationError) -> str:
    try:
        return json.dumps(exc.errors(), ensure_ascii=False)[:8000]
    except Exception:
        return str(exc)[:8000]


class NormExtractionService:
    """Orchestrate LLM extraction, validation, NormCreate import, and artifact write."""

    def __init__(self, norm_service: NormService | None = None):
        self._norm_service = norm_service or NormService()

    async def extract_from_text(
        self,
        db: AsyncSession,
        *,
        text: str,
        default_lang: str,
        field_id: Optional[UUID],
        created_by: Optional[str],
        max_norms: int,
        default_authority_level: Optional[str],
        default_formality_level: Optional[str],
        submitter_user_id: Optional[str],
        llm_doc_id: Optional[str] = None,
        llm_doc_title: Optional[str] = None,
        source_extraction_extra: Optional[Dict[str, Any]] = None,
        artifact_relative_segments: Optional[tuple[str, ...]] = None,
    ) -> Tuple[UUID, str, List[UUID], List[Dict[str, Any]]]:
        session_id = uuid.uuid4()
        artifact_mode = "inline_text"
        if source_extraction_extra and source_extraction_extra.get("mode"):
            artifact_mode = str(source_extraction_extra["mode"])
        doc_id = llm_doc_id or "inline"
        doc_title = llm_doc_title or "inline-text"

        try:
            raw_out = await run_raw_extraction(
                text or "",
                default_lang=default_lang,
                field_id=field_id,
                created_by=created_by,
                max_norms=max_norms,
                default_authority_level=default_authority_level,
                default_formality_level=default_formality_level,
                doc_id=doc_id,
                doc_title=doc_title,
            )
        except LlmExtractionError as e:
            raise HTTPException(
                status_code=status.HTTP_502_BAD_GATEWAY,
                detail=f"LLM extraction failed: {e.message}",
            ) from e
        except BadExtractionJsonError as e:
            raise HTTPException(
                status_code=status.HTTP_502_BAD_GATEWAY,
                detail=e.message,
            ) from e

        norms_raw = raw_out.norms_raw
        chunk_results = raw_out.chunk_results
        chunks_len = raw_out.chunk_count
        max_chars = raw_out.max_content_chars
        overlap = raw_out.overlap_chars

        norm_ids: List[UUID] = []
        issues: List[Dict[str, Any]] = []
        artifact_norms: List[Dict[str, Any]] = []

        for idx, item in enumerate(norms_raw):
            if not isinstance(item, dict):
                issues.append({"index": idx, "detail": "Item is not an object."})
                artifact_norms.append(
                    {
                        "index": idx,
                        "status": "skipped",
                        "validation_errors": ["not an object"],
                    }
                )
                continue
            coerced = _coerce_norm_dict(item)
            if field_id is not None and coerced.get("field_id") is None:
                coerced["field_id"] = field_id
            if created_by is not None and coerced.get("created_by") is None:
                coerced["created_by"] = created_by
            if coerced.get("authority_level") is None and default_authority_level:
                coerced["authority_level"] = default_authority_level
            if coerced.get("formality_level") is None and default_formality_level:
                coerced["formality_level"] = default_formality_level

            sd = coerced.get("source_details")
            if not isinstance(sd, dict):
                sd = {}
            ext = sd.get("extraction") if isinstance(sd.get("extraction"), dict) else {}
            ext = {
                **ext,
                "mode": "inline_text",
                "session_id": str(session_id),
            }
            if chunks_len > 1:
                ext["text_chunk_count"] = chunks_len
            if submitter_user_id:
                ext["submitter_user_id"] = submitter_user_id
            if source_extraction_extra:
                ext = {**ext, **source_extraction_extra}
            sd = {**sd, "extraction": ext}
            coerced["source_details"] = sd

            try:
                norm_create = NormCreate.model_validate(coerced)
            except ValidationError as e:
                issues.append({"index": idx, "detail": _validation_detail(e)})
                artifact_norms.append(
                    {
                        "index": idx,
                        "status": "validation_failed",
                        "payload": coerced,
                        "validation_errors": e.errors(),
                    }
                )
                continue

            try:
                created = await self._norm_service.create_norm(db, norm_create)
            except Exception as e:
                issues.append({"index": idx, "detail": str(e)[:8000]})
                artifact_norms.append(
                    {
                        "index": idx,
                        "status": "import_failed",
                        "payload": coerced,
                        "detail": str(e),
                    }
                )
                continue

            norm_ids.append(created.id)
            artifact_norms.append(
                {
                    "index": idx,
                    "status": "imported",
                    "norm_id": str(created.id),
                }
            )

        if chunks_len <= 1:
            single_raw = chunk_results[0]["raw_llm"] if chunk_results else ""
            artifact_body: Dict[str, Any] = {
                "schema_version": 1,
                "session_id": str(session_id),
                "mode": artifact_mode,
                "raw_llm": single_raw,
                "norms": artifact_norms,
            }
        else:
            artifact_body = {
                "schema_version": 2,
                "session_id": str(session_id),
                "mode": artifact_mode,
                "chunking": {
                    "enabled": True,
                    "max_content_chars": max_chars,
                    "overlap_chars": overlap,
                    "chunk_count": chunks_len,
                },
                "chunk_results": chunk_results,
                "norms": artifact_norms,
            }
        rel_segments: tuple[str, ...] = artifact_relative_segments or (
            "inline",
            str(session_id),
            "process.json",
        )
        rel = write_process_json(
            artifact_root=settings.extraction_artifact_dir,
            relative_segments=rel_segments,
            data=artifact_body,
            max_bytes=settings.extraction_artifact_max_bytes,
        )

        return session_id, rel, norm_ids, issues



"""Compliance check and outcome reporting APIs."""
from __future__ import annotations

from fastapi import APIRouter, Depends, Header
from sqlalchemy.ext.asyncio import AsyncSession

from normbase.db.connection import get_db
from normbase.schemas import (
    ComplianceCheckRequest,
    ComplianceCheckResponse,
    NormOutcomeReportRequest,
    NormOutcomeReportResponse,
)
from normbase.services.compliance_service import ComplianceService

router = APIRouter(prefix="/norms", tags=["norms"])
_compliance_service = ComplianceService()


@router.post("/check_action_compliance", response_model=ComplianceCheckResponse)
async def check_action_compliance(
    request: ComplianceCheckRequest,
    db: AsyncSession = Depends(get_db),
    x_agent_id: str | None = Header(default=None),
    x_key_id: str | None = Header(default=None),
):
    return await _compliance_service.check_action_compliance(
        db,
        request=request,
        x_agent_id=x_agent_id,
        x_key_id=x_key_id,
    )


@router.post("/log-outcome", response_model=NormOutcomeReportResponse)
async def report_norm_outcome(
    request: NormOutcomeReportRequest,
    db: AsyncSession = Depends(get_db),
    x_agent_id: str | None = Header(default=None),
    x_key_id: str | None = Header(default=None),
):
    return await _compliance_service.report_norm_outcome(
        db,
        request=request,
        x_agent_id=x_agent_id,
        x_key_id=x_key_id,
    )



"""Normbase client abstraction and adapters for EDA engine."""
from __future__ import annotations

from typing import Any, Dict, List, Optional, Protocol, Literal

import aiohttp

from security.auth.verifier import build_internal_auth_headers


class NormbaseClient(Protocol):
    """Abstract interface for EDA to consume norm retrieval services."""

    async def retrieve_norms(
        self,
        context: Dict[str, Any],
        query_mode: Literal["structured", "semantic", "hybrid"] = "hybrid",
        nl_query: Optional[str] = None,
        agent_role: Optional[str] = None,
        agent_ids: Optional[List[str]] = None,
        tag_filters: Optional[Dict[str, List[str]]] = None,
        field_ids: Optional[List[str]] = None,
        top_k: int = 5,
    ) -> List[Dict[str, Any]]:
        ...

    async def retrieve_norms_with_meta(
        self,
        context: Dict[str, Any],
        query_mode: Literal["structured", "semantic", "hybrid"] = "hybrid",
        nl_query: Optional[str] = None,
        agent_role: Optional[str] = None,
        agent_ids: Optional[List[str]] = None,
        tag_filters: Optional[Dict[str, List[str]]] = None,
        field_ids: Optional[List[str]] = None,
        top_k: int = 5,
    ) -> Dict[str, Any]:
        ...

    async def get_default_norms(self, agent_role: str) -> List[Dict[str, Any]]:
        ...

    async def report_outcome(
        self, agent_id: str, norm_id: str, outcome: str, details: Dict[str, Any]
    ) -> None:
        ...


class MockNormbaseClient:
    """Simple in-memory normbase adapter for local prototype tests."""

    def __init__(self, norms: Optional[List[Dict[str, Any]]] = None):
        self.norms = norms or []

    async def retrieve_norms(
        self,
        context: Dict[str, Any],
        query_mode: Literal["structured", "semantic", "hybrid"] = "hybrid",
        nl_query: Optional[str] = None,
        agent_role: Optional[str] = None,
        agent_ids: Optional[List[str]] = None,
        tag_filters: Optional[Dict[str, List[str]]] = None,
        field_ids: Optional[List[str]] = None,
        top_k: int = 5,
    ) -> List[Dict[str, Any]]:
        if query_mode == "semantic" and not str(nl_query or "").strip():
            raise ValueError("nl_query is required when query_mode='semantic'.")
        del context, field_ids, nl_query
        effective_tag_filters: Dict[str, List[str]] = {}
        if isinstance(tag_filters, dict):
            for category, raw_values in tag_filters.items():
                key = str(category).strip()
                if not key:
                    continue
                if not isinstance(raw_values, list):
                    continue
                values = [str(v).strip() for v in raw_values if str(v).strip()]
                if values:
                    effective_tag_filters[key] = values
        if agent_role:
            role_values = effective_tag_filters.get("role", [])
            if agent_role not in role_values:
                effective_tag_filters["role"] = [*role_values, agent_role]
        normalized_agent_ids = sorted(
            {
                str(agent_id).strip()
                for agent_id in (agent_ids or [])
                if str(agent_id).strip()
            }
        )

        def _has_tag(norm: Dict[str, Any], category: str, values: List[str]) -> bool:
            if not values:
                return True
            expected_values = list(values)
            if category == "role" and "*" not in expected_values:
                # role='*' means global role rule (applies to everyone)
                expected_values.append("*")
            tags = norm.get("tags", []) or []
            for tag in tags:
                if not isinstance(tag, dict):
                    continue
                if str(tag.get("tag_category", "")).strip() != category:
                    continue
                if str(tag.get("tag", "")).strip() in expected_values:
                    return True
            return False

        def _passes_filters(norm: Dict[str, Any]) -> bool:
            if normalized_agent_ids:
                subject = norm.get("subject", {}) if isinstance(norm, dict) else {}
                canonical = (
                    subject.get("canonical", {}) if isinstance(subject, dict) else {}
                )
                raw_agents = canonical.get("agents", []) if isinstance(canonical, dict) else []
                norm_agents = {
                    str(item).strip()
                    for item in (raw_agents if isinstance(raw_agents, list) else [])
                    if str(item).strip()
                }
                if not norm_agents.intersection(normalized_agent_ids):
                    return False
            for category, values in effective_tag_filters.items():
                if category == "role":
                    if _has_tag(norm, "role", values):
                        continue
                    return False
                if not _has_tag(norm, category, values):
                    return False
            return True

        scoped = [
            n
            for n in self.norms
            if _passes_filters(n) and n.get("status", "active") == "active"
        ]
        return scoped[:top_k]

    async def retrieve_norms_with_meta(
        self,
        context: Dict[str, Any],
        query_mode: Literal["structured", "semantic", "hybrid"] = "hybrid",
        nl_query: Optional[str] = None,
        agent_role: Optional[str] = None,
        agent_ids: Optional[List[str]] = None,
        tag_filters: Optional[Dict[str, List[str]]] = None,
        field_ids: Optional[List[str]] = None,
        top_k: int = 5,
    ) -> Dict[str, Any]:
        norms = await self.retrieve_norms(
            context=context,
            query_mode=query_mode,
            nl_query=nl_query,
            agent_role=agent_role,
            agent_ids=agent_ids,
            tag_filters=tag_filters,
            field_ids=field_ids,
            top_k=top_k,
        )
        return {
            "norms": norms,
            "scores": [1.0 for _ in norms],
            "total": len(norms),
            "usage_token": "",
            "cache_hit": False,
            "meta": {
                "query_mode_effective": query_mode,
                "semantic_used": False,
                "semantic_attempted": bool(str(nl_query or "").strip()),
                "semantic_query_source": "nl_query" if str(nl_query or "").strip() else "none",
                "sql_candidates": len(norms),
                "vector_hits": 0,
                "semantic_hits": 0,
                "fusion_strategy": "structured_only",
            },
        }

    async def get_default_norms(self, agent_role: str) -> List[Dict[str, Any]]:
        return await self.retrieve_norms(context={}, agent_role=agent_role, top_k=20)

    async def report_outcome(
        self, agent_id: str, norm_id: str, outcome: str, details: Dict[str, Any]
    ) -> None:
        del agent_id, norm_id, outcome, details
        return None


class HttpNormbaseClient:
    """HTTP adapter over the existing Normbase API."""

    def __init__(
        self,
        base_url: str,
        retrieve_timeout_seconds: int = 60,
        outcome_timeout_seconds: int = 20,
    ):
        self.base_url = base_url.rstrip("/")
        self.retrieve_timeout_seconds = retrieve_timeout_seconds
        self.outcome_timeout_seconds = outcome_timeout_seconds

    async def retrieve_norms(
        self,
        context: Dict[str, Any],
        query_mode: Literal["structured", "semantic", "hybrid"] = "hybrid",
        nl_query: Optional[str] = None,
        agent_role: Optional[str] = None,
        agent_ids: Optional[List[str]] = None,
        tag_filters: Optional[Dict[str, List[str]]] = None,
        field_ids: Optional[List[str]] = None,
        top_k: int = 5,
    ) -> List[Dict[str, Any]]:
        data = await self.retrieve_norms_with_meta(
            context=context,
            query_mode=query_mode,
            nl_query=nl_query,
            agent_role=agent_role,
            agent_ids=agent_ids,
            tag_filters=tag_filters,
            field_ids=field_ids,
            top_k=top_k,
        )
        return list(data.get("norms", []) or [])

    async def retrieve_norms_with_meta(
        self,
        context: Dict[str, Any],
        query_mode: Literal["structured", "semantic", "hybrid"] = "hybrid",
        nl_query: Optional[str] = None,
        agent_role: Optional[str] = None,
        agent_ids: Optional[List[str]] = None,
        tag_filters: Optional[Dict[str, List[str]]] = None,
        field_ids: Optional[List[str]] = None,
        top_k: int = 5,
    ) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "context": context,
            "query_mode": query_mode,
            "top_k": top_k,
        }
        if nl_query:
            payload["nl_query"] = nl_query
        if agent_role:
            payload["agent_role"] = agent_role
        if agent_ids:
            payload["agent_ids"] = agent_ids
        if tag_filters:
            payload["tag_filters"] = tag_filters
        if field_ids:
            payload["field_ids"] = field_ids
        headers = build_internal_auth_headers()
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{self.base_url}/api/v1/retrieve",
                json=payload,
                headers=headers,
                timeout=aiohttp.ClientTimeout(total=self.retrieve_timeout_seconds),
            ) as response:
                response.raise_for_status()
                data = await response.json()
                return data or {}

    async def get_default_norms(self, agent_role: str) -> List[Dict[str, Any]]:
        # Prototype default: use retrieval with empty context.
        return await self.retrieve_norms(context={}, agent_role=agent_role, top_k=20)

    async def report_outcome(
        self, agent_id: str, norm_id: str, outcome: str, details: Dict[str, Any]
    ) -> None:
        usage_token = str((details or {}).get("usage_token", "")).strip()
        if not usage_token:
            # Normbase log-outcome requires usage_token binding; keep this best-effort
            # API as a no-op unless caller supplies it in details.
            return None
        payload = {
            "usage_token": usage_token,
            "agent_id": agent_id,
            "norm_id": norm_id,
            "outcome": outcome,
            "details": details,
        }
        headers = build_internal_auth_headers()
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{self.base_url}/api/v1/norms/log-outcome",
                json=payload,
                headers=headers,
                timeout=aiohttp.ClientTimeout(total=self.outcome_timeout_seconds),
            ) as response:
                if response.status >= 400:
                    return None
        return None



"""Shared auth core for multi-entry integrations."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Literal, Sequence
from uuid import UUID
import os

import httpx

from security.auth.api_keys import ResolvedAPIKey

EntryType = Literal["mcp", "openclaw", "sdk"]
AuthDecision = Literal["allow", "deny"]

REASON_OK = "ok"
REASON_MISSING_API_KEY = "missing_api_key"
REASON_INVALID_API_KEY = "invalid_api_key"
REASON_INACTIVE_KEY = "inactive_key"
REASON_EXPIRED_KEY = "expired_key"
REASON_MISSING_SCOPE = "missing_scope"
REASON_AGENT_MISMATCH = "agent_mismatch"

SCOPE_WILDCARD = "*"


def _normbase_base_url() -> str:
    return os.getenv("NORMBASE_BASE_URL", "http://localhost:8001").rstrip("/")


@dataclass(frozen=True)
class AuthInput:
    """Normalized authorization input across all integration entries."""

    entry_type: EntryType
    api_key: str
    action: str
    required_scopes: list[str]
    resource_agent_id: str | None = None
    request_id: str | None = None
    caller_id: str | None = None


@dataclass(frozen=True)
class AuthContext:
    """Authorization decision and resolved identity context."""

    decision: AuthDecision
    reason: str
    agent_id: str
    key_id: str
    scopes: list[str]
    request_id: str | None
    caller_id: str | None
    action: str


def _normalize_scope_list(value: Sequence[str] | None) -> list[str]:
    scopes: list[str] = []
    for raw in value or []:
        text = str(raw).strip()
        if text and text not in scopes:
            scopes.append(text)
    return scopes


def _normalize_agent_id_or_empty(raw_agent_id: str | None) -> str:
    text = (raw_agent_id or "").strip()
    if not text:
        return ""
    try:
        return str(UUID(text))
    except ValueError:
        return ""


def _utc_now() -> datetime:
    return datetime.now(timezone.utc)


def _normalize_expires_at(value: object) -> datetime | None:
    if value is None:
        return None
    if isinstance(value, datetime):
        if value.tzinfo is None:
            return value.replace(tzinfo=timezone.utc)
        return value.astimezone(timezone.utc)
    if isinstance(value, str):
        raw = value.strip()
        if not raw:
            return None
        candidate = raw.replace("Z", "+00:00")
        try:
            parsed = datetime.fromisoformat(candidate)
        except ValueError:
            return None
        if parsed.tzinfo is None:
            return parsed.replace(tzinfo=timezone.utc)
        return parsed.astimezone(timezone.utc)
    return None


def _has_scope(scopes: Sequence[str], required_scope: str) -> bool:
    normalized = _normalize_scope_list(scopes)
    if SCOPE_WILDCARD in normalized:
        return True
    return required_scope in normalized


async def resolve_agent_api_key(api_key: str) -> ResolvedAPIKey | None:
    """Resolve per-agent API key via runtime introspection endpoint."""
    key = (api_key or "").strip()
    if not key:
        return None
    headers = {"X-API-Key": key}
    try:
        async with httpx.AsyncClient(timeout=20.0) as client:
            resp = await client.get(
                f"{_normbase_base_url()}/api/v1/keys/introspect",
                headers=headers,
            )
            if resp.status_code >= 400:
                return None
            data = resp.json() or {}
            return ResolvedAPIKey(
                key_id=str(data.get("key_id", "")).strip(),
                agent_id=str(data.get("agent_id", "")).strip(),
                scopes=_normalize_scope_list(data.get("scopes", [SCOPE_WILDCARD])),
                is_active=bool(data.get("is_active", True)),
                expires_at=_normalize_expires_at(data.get("expires_at")),
            )
    except Exception:
        return None


async def authorize(
    auth_input: AuthInput,
    *,
    resolver=resolve_agent_api_key,
) -> AuthContext:
    """Authorize one integration action with consistent decision semantics."""
    api_key = (auth_input.api_key or "").strip()
    if not api_key:
        return AuthContext(
            decision="deny",
            reason=REASON_MISSING_API_KEY,
            agent_id="",
            key_id="",
            scopes=[],
            request_id=auth_input.request_id,
            caller_id=auth_input.caller_id,
            action=auth_input.action,
        )

    resolved = await resolver(api_key)
    if resolved is None or not str(resolved.agent_id or "").strip():
        return AuthContext(
            decision="deny",
            reason=REASON_INVALID_API_KEY,
            agent_id="",
            key_id="",
            scopes=[],
            request_id=auth_input.request_id,
            caller_id=auth_input.caller_id,
            action=auth_input.action,
        )
    if not resolved.is_active:
        return AuthContext(
            decision="deny",
            reason=REASON_INACTIVE_KEY,
            agent_id="",
            key_id=resolved.key_id,
            scopes=_normalize_scope_list(resolved.scopes),
            request_id=auth_input.request_id,
            caller_id=auth_input.caller_id,
            action=auth_input.action,
        )

    expires_at = _normalize_expires_at(resolved.expires_at)
    if expires_at is not None and expires_at <= _utc_now():
        return AuthContext(
            decision="deny",
            reason=REASON_EXPIRED_KEY,
            agent_id="",
            key_id=resolved.key_id,
            scopes=_normalize_scope_list(resolved.scopes),
            request_id=auth_input.request_id,
            caller_id=auth_input.caller_id,
            action=auth_input.action,
        )

    required_scopes = _normalize_scope_list(auth_input.required_scopes)
    principal_scopes = _normalize_scope_list(resolved.scopes)
    for scope in required_scopes:
        if _has_scope(principal_scopes, scope):
            continue
        return AuthContext(
            decision="deny",
            reason=f"{REASON_MISSING_SCOPE}:{scope}",
            agent_id=_normalize_agent_id_or_empty(resolved.agent_id),
            key_id=resolved.key_id,
            scopes=principal_scopes,
            request_id=auth_input.request_id,
            caller_id=auth_input.caller_id,
            action=auth_input.action,
        )

    resolved_agent_id = _normalize_agent_id_or_empty(resolved.agent_id)
    resource_agent_id = _normalize_agent_id_or_empty(auth_input.resource_agent_id)
    if resource_agent_id and resolved_agent_id != resource_agent_id:
        return AuthContext(
            decision="deny",
            reason=REASON_AGENT_MISMATCH,
            agent_id=resolved_agent_id,
            key_id=resolved.key_id,
            scopes=principal_scopes,
            request_id=auth_input.request_id,
            caller_id=auth_input.caller_id,
            action=auth_input.action,
        )

    return AuthContext(
        decision="allow",
        reason=REASON_OK,
        agent_id=resolved_agent_id,
        key_id=resolved.key_id,
        scopes=principal_scopes,
        request_id=auth_input.request_id,
        caller_id=auth_input.caller_id,
        action=auth_input.action,
    )



"""Application service for EDA agent CRUD and state mapping."""
from __future__ import annotations

from datetime import datetime
from typing import List
import uuid

from fastapi import HTTPException
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from eda_agent.models.persistence import AgentProfile, AgentStateSnapshot
from eda_agent.schemas import (
    AgentCreateRequest,
    AgentListItemResponse,
    AgentStateResponse,
    AgentSummaryResponse,
    AgentUpdateRequest,
)
from eda_agent.services.field_service import FieldService
from eda_agent.state import EDAState, ValueWeight


class AgentService:
    """Service layer for profile and snapshot lifecycle."""

    def __init__(self, field_service: FieldService | None = None):
        self.field_service = field_service or FieldService()

    async def get_profile_and_snapshot(
        self, db: AsyncSession, agent_id: uuid.UUID
    ) -> tuple[AgentProfile, AgentStateSnapshot]:
        profile = (
            await db.execute(select(AgentProfile).where(AgentProfile.id == agent_id))
        ).scalar_one_or_none()
        if not profile:
            raise HTTPException(status_code=404, detail=f"Agent {agent_id} not found")

        snapshot = (
            await db.execute(
                select(AgentStateSnapshot).where(AgentStateSnapshot.agent_id == agent_id)
            )
        ).scalar_one_or_none()
        if not snapshot:
            raise HTTPException(
                status_code=500, detail=f"State snapshot missing for agent {agent_id}"
            )
        return profile, snapshot

    def state_from_snapshot(
        self, profile: AgentProfile, snapshot: AgentStateSnapshot
    ) -> EDAState:
        payload = {
            "agent_id": str(profile.id),
            "agent_role": profile.agent_role,
            "agent_name": profile.agent_name,
            "beliefs": snapshot.beliefs or {},
            "norm_bindings": snapshot.norm_bindings or [],
            "value_weights": snapshot.value_weights or [],
            "active_fields": snapshot.active_fields or [],
            "current_intention": snapshot.current_intention,
            "trust_score": snapshot.trust_score,
            "created_at": snapshot.created_at.isoformat(),
            "updated_at": snapshot.updated_at.isoformat(),
        }
        return EDAState.from_dict(payload)

    def state_to_response(
        self, state: EDAState, profile: AgentProfile, snapshot: AgentStateSnapshot
    ) -> AgentStateResponse:
        data = state.to_dict()
        return AgentStateResponse(
            id=profile.id,
            agent_name=data.get("agent_name") or profile.agent_name,
            agent_role=data["agent_role"],
            agent_description=profile.agent_description or "",
            org_id=profile.org_id,
            beliefs=data["beliefs"],
            norm_bindings=data["norm_bindings"],
            value_weights=data["value_weights"],
            active_fields=data["active_fields"],
            current_intention=data["current_intention"],
            trust_score=data["trust_score"],
            created_at=snapshot.created_at,
            updated_at=snapshot.updated_at,
        )

    def profile_snapshot_to_list_item(
        self, profile: AgentProfile, snapshot: AgentStateSnapshot | None
    ) -> AgentListItemResponse:
        beliefs = snapshot.beliefs if snapshot and isinstance(snapshot.beliefs, dict) else {}
        norm_bindings = (
            snapshot.norm_bindings if snapshot and isinstance(snapshot.norm_bindings, list) else []
        )
        value_weights = (
            snapshot.value_weights if snapshot and isinstance(snapshot.value_weights, list) else []
        )
        active_fields = (
            snapshot.active_fields if snapshot and isinstance(snapshot.active_fields, list) else []
        )
        current_intention = snapshot.current_intention if snapshot else None
        trust_score = float(snapshot.trust_score) if snapshot else 0.8
        created_at = snapshot.created_at if snapshot else profile.created_at
        updated_at = snapshot.updated_at if snapshot else profile.updated_at
        return AgentListItemResponse(
            id=profile.id,
            agent_name=profile.agent_name,
            agent_role=profile.agent_role,
            agent_description=profile.agent_description or "",
            org_id=profile.org_id,
            beliefs=beliefs,
            norm_bindings=norm_bindings,
            value_weights=value_weights,
            active_fields=active_fields,
            current_intention=current_intention,
            trust_score=trust_score,
            created_at=created_at,
            updated_at=updated_at,
        )

    async def create_agent(self, db: AsyncSession, request: AgentCreateRequest) -> AgentSummaryResponse:
        exists = (
            await db.execute(select(AgentProfile).where(AgentProfile.agent_name == request.agent_name))
        ).scalar_one_or_none()
        if exists:
            raise HTTPException(status_code=409, detail=f"Agent {request.agent_name} already exists")

        if "active_fields" in request.model_fields_set:
            resolved_active_fields = await self.field_service.validate_active_fields(
                db, request.active_fields
            )
        else:
            resolved_active_fields = await self.field_service.resolve_active_fields(
                db, request.active_fields
            )

        profile = AgentProfile(
            agent_name=request.agent_name,
            agent_role=request.agent_role,
            agent_description=request.agent_description or "",
            org_id=request.org_id,
        )
        db.add(profile)

        await db.flush()
        value_weights = [
            ValueWeight(
                value_name=i.value_name,
                weight=i.weight,
                hierarchy=i.hierarchy,
                rationale=i.rationale,
            )
            for i in request.value_weights
        ]
        initial_state = EDAState(
            agent_id=str(profile.id),
            agent_role=request.agent_role,
            agent_name=request.agent_name,
            value_weights=value_weights,
            active_fields=resolved_active_fields,
            trust_score=request.trust_score,
        )
        state_dict = initial_state.to_dict()
        snapshot = AgentStateSnapshot(
            agent_id=profile.id,
            beliefs=state_dict["beliefs"],
            norm_bindings=state_dict["norm_bindings"],
            value_weights=state_dict["value_weights"],
            active_fields=state_dict["active_fields"],
            current_intention=initial_state.current_intention,
            trust_score=initial_state.trust_score,
        )
        db.add(snapshot)
        await db.commit()
        await db.refresh(profile)
        return AgentSummaryResponse.model_validate(profile)

    async def list_agents(
        self, db: AsyncSession, *, limit: int
    ) -> List[AgentListItemResponse]:
        rows = (
            await db.execute(
                select(AgentProfile, AgentStateSnapshot)
                .join(
                    AgentStateSnapshot,
                    AgentStateSnapshot.agent_id == AgentProfile.id,
                    isouter=True,
                )
                .order_by(AgentProfile.updated_at.desc())
                .limit(limit)
            )
        ).all()
        return [
            self.profile_snapshot_to_list_item(profile, snapshot)
            for profile, snapshot in rows
        ]

    async def get_agent_state(
        self, db: AsyncSession, *, agent_id: uuid.UUID
    ) -> AgentStateResponse:
        profile, snapshot = await self.get_profile_and_snapshot(db, agent_id)
        state = self.state_from_snapshot(profile, snapshot)
        return self.state_to_response(state, profile, snapshot)

    async def update_agent(
        self,
        db: AsyncSession,
        *,
        agent_id: uuid.UUID,
        request: AgentUpdateRequest,
    ) -> AgentSummaryResponse:
        profile, snapshot = await self.get_profile_and_snapshot(db, agent_id)
        update_data = request.model_dump(exclude_unset=True)

        if "agent_role" in update_data:
            profile.agent_role = update_data["agent_role"]
        if "agent_description" in update_data and update_data["agent_description"] is not None:
            profile.agent_description = update_data["agent_description"]
        if "org_id" in update_data:
            profile.org_id = update_data["org_id"]

        if "active_fields" in update_data:
            snapshot.active_fields = await self.field_service.validate_active_fields(
                db, update_data["active_fields"]
            )
        if "trust_score" in update_data and update_data["trust_score"] is not None:
            snapshot.trust_score = update_data["trust_score"]

        profile.updated_at = datetime.now()
        snapshot.updated_at = datetime.now()
        await db.commit()
        await db.refresh(profile)
        return AgentSummaryResponse.model_validate(profile)

    async def delete_agent(self, db: AsyncSession, *, agent_id: uuid.UUID) -> None:
        profile, _snapshot = await self.get_profile_and_snapshot(db, agent_id)
        await db.delete(profile)
        await db.commit()
